# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import logging
import os
import struct
import sys
import unittest

try:
    import mock  # Python 2
except ImportError:
    from unittest import mock  # Python 3

from nose.tools import eq_
from nose.tools import raises

from ryu.utils import binary_str
from ryu.lib import pcaplib

LOG = logging.getLogger(__name__)

PCAP_PACKET_DATA_DIR = os.path.join(
    os.path.dirname(sys.modules[__name__].__file__),
    '../../packet_data/pcap/')


class Test_PcapFileHdr(unittest.TestCase):
    """
    Test case for pcaplib.PcapFileHdr class
    """
    hdr = pcaplib.PcapFileHdr(
        magic=None,  # temporary default
        version_major=2,
        version_minor=4,
        thiszone=0x11223344,
        sigfigs=0x22334455,
        snaplen=0x33445566,
        network=0x44556677,
    )

    buf_big = (
        b'\xa1\xb2\xc3\xd4'  # magic (Big Endian)
        b'\x00\x02\x00\x04'  # version_major, version_minor
        b'\x11\x22\x33\x44'  # thiszone
        b'\x22\x33\x44\x55'  # sigfigs
        b'\x33\x44\x55\x66'  # snaplen
        b'\x44\x55\x66\x77'  # network
    )

    buf_little = (
        b'\xd4\xc3\xb2\xa1'  # magic (Little Endian)
        b'\x02\x00\x04\x00'  # version_major, version_minor
        b'\x44\x33\x22\x11'  # thiszone
        b'\x55\x44\x33\x22'  # sigfigs
        b'\x66\x55\x44\x33'  # snaplen
        b'\x77\x66\x55\x44'  # network
    )

    buf_invalid = (
        b'\xff\xff\xff\xff'  # magic (Invalid)
        b'\x02\x00\x04\x00'  # version_major, version_minor
        b'\x44\x33\x22\x11'  # thiszone
        b'\x55\x44\x33\x22'  # sigfigs
        b'\x66\x55\x44\x33'  # snaplen
        b'\x77\x66\x55\x44'  # network
    )

    def _assert(self, magic, ret):
        self.hdr.magic = magic
        eq_(self.hdr.__dict__, ret.__dict__)

    def test_parser_with_big_endian(self):
        ret, byteorder = pcaplib.PcapFileHdr.parser(self.buf_big)
        self._assert(pcaplib.PcapFileHdr.MAGIC_NUMBER_IDENTICAL, ret)
        eq_('big', byteorder)

    def test_parser_with_little_endian(self):
        ret, byteorder = pcaplib.PcapFileHdr.parser(self.buf_little)
        self._assert(pcaplib.PcapFileHdr.MAGIC_NUMBER_SWAPPED, ret)
        eq_('little', byteorder)

    @mock.patch('sys.byteorder', 'big')
    def test_serialize_with_big_endian(self):
        buf = self.hdr.serialize()
        eq_(binary_str(self.buf_big), binary_str(buf))

    @mock.patch('sys.byteorder', 'little')
    def test_serialize_with_little_endian(self):
        buf = self.hdr.serialize()
        eq_(binary_str(self.buf_little), binary_str(buf))

    @raises(struct.error)
    def test_parser_with_invalid_magic_number(self):
        pcaplib.PcapFileHdr.parser(self.buf_invalid)


class Test_PcapPktHdr(unittest.TestCase):
    """
    Test case for pcaplib.PcapPktHdr class
    """
    expected_buf = b'test_data'

    hdr = pcaplib.PcapPktHdr(
        ts_sec=0x11223344,
        ts_usec=0x22334455,
        incl_len=len(expected_buf),
        orig_len=0x44556677,
    )

    buf_big = (
        b'\x11\x22\x33\x44'  # ts_sec
        b'\x22\x33\x44\x55'  # ts_usec
        b'\x00\x00\x00\x09'  # incl_len = len(expected_buf)
        b'\x44\x55\x66\x77'  # orig_len
    )

    buf_little = (
        b'\x44\x33\x22\x11'  # ts_sec
        b'\x55\x44\x33\x22'  # ts_usec
        b'\x09\x00\x00\x00'  # incl_len = len(expected_buf)
        b'\x77\x66\x55\x44'  # orig_len
    )

    def test_parser_with_big_endian(self):
        ret, buf = pcaplib.PcapPktHdr.parser(
            self.buf_big + self.expected_buf, 'big')
        eq_(self.hdr.__dict__, ret.__dict__)
        eq_(self.expected_buf, buf)

    def test_parser_with_little_endian(self):
        ret, buf = pcaplib.PcapPktHdr.parser(
            self.buf_little + self.expected_buf, 'little')
        eq_(self.hdr.__dict__, ret.__dict__)
        eq_(self.expected_buf, buf)

    @mock.patch('sys.byteorder', 'big')
    def test_serialize_with_big_endian(self):
        buf = self.hdr.serialize()
        eq_(binary_str(self.buf_big), binary_str(buf))

    @mock.patch('sys.byteorder', 'little')
    def test_serialize_with_little_endian(self):
        buf = self.hdr.serialize()
        eq_(binary_str(self.buf_little), binary_str(buf))


class Test_pcaplib_Reader(unittest.TestCase):
    """
    Test case for pcaplib.Reader class
    """

    expected_outputs = [
        (0x1234 + (0x5678 / 1e6), b'test_data_1'),  # sec=0x1234, usec=0x5678
        (0x2345 + (0x6789 / 1e6), b'test_data_2'),  # sec=0x2345, usec=0x6789
    ]

    def _test(self, file_name):
        outputs = []
        for ts, buf in pcaplib.Reader(open(file_name, 'rb')):
            outputs.append((ts, buf))

        eq_(self.expected_outputs, outputs)

    def test_with_big_endian(self):
        self._test(os.path.join(PCAP_PACKET_DATA_DIR, 'big_endian.pcap'))

    def test_with_little_endian(self):
        self._test(os.path.join(PCAP_PACKET_DATA_DIR, 'little_endian.pcap'))


class DummyFile(object):

    def __init__(self):
        self.buf = b''

    def write(self, buf):
        self.buf += buf

    def close(self):
        pass


class Test_pcaplib_Writer(unittest.TestCase):
    """
    Test case for pcaplib.Writer class
    """

    @staticmethod
    def _test(file_name):
        expected_buf = open(file_name, 'rb').read()
        f = DummyFile()
        w = pcaplib.Writer(f)
        w.write_pkt(b'test_data_1', ts=(0x1234 + (0x5678 / 1e6)))
        w.write_pkt(b'test_data_2', ts=(0x2345 + (0x6789 / 1e6)))
        eq_(expected_buf, f.buf)

    @mock.patch('sys.byteorder', 'big')
    def test_with_big_endian(self):
        self._test(os.path.join(PCAP_PACKET_DATA_DIR, 'big_endian.pcap'))

    @mock.patch('sys.byteorder', 'little')
    def test_with_little_endian(self):
        self._test(os.path.join(PCAP_PACKET_DATA_DIR, 'little_endian.pcap'))

    @staticmethod
    @mock.patch.object(pcaplib.Writer, '_write_pcap_file_hdr', mock.MagicMock)
    @mock.patch.object(pcaplib.Writer, '_write_pkt_hdr', mock.MagicMock)
    def test_with_longer_buf():
        f = DummyFile()
        snaplen = 4
        w = pcaplib.Writer(f, snaplen=snaplen)
        w.write_pkt(b'hogehoge', ts=0)
        expected_buf = b'hoge'  # b'hogehoge'[:snaplen]
        eq_(expected_buf, f.buf)
        eq_(snaplen, len(f.buf))



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Api for operator. Mainly commands to build CLI and
 operator interface around them.
"""
import logging

from ryu.services.protocols.bgp.api.base import ApiException
from ryu.services.protocols.bgp.api.base import register
from ryu.services.protocols.bgp.api.rpc_log_handler import RpcLogHandler
from ryu.services.protocols.bgp.operator.command import Command
from ryu.services.protocols.bgp.operator.command import STATUS_ERROR
from ryu.services.protocols.bgp.operator.commands.clear import ClearCmd
from ryu.services.protocols.bgp.operator.commands.set import SetCmd
from ryu.services.protocols.bgp.operator.commands.show import ShowCmd
from ryu.services.protocols.bgp.operator.internal_api import InternalApi

LOG = logging.getLogger('bgpspeaker.api.rtconf')

DEFAULT_LOG_FORMAT = '%(asctime)s %(levelname)s %(message)s'


def _init_log_handler():
    log_handler = RpcLogHandler()
    log_handler.setLevel(logging.ERROR)
    log_handler.formatter = logging.Formatter(DEFAULT_LOG_FORMAT)
    return log_handler


INTERNAL_API = InternalApi(_init_log_handler())


class RootCmd(Command):
    subcommands = {
        'show': ShowCmd,
        'set': SetCmd,
        'clear': ClearCmd}


def operator_run(cmd, **kwargs):
    params = kwargs.get('params', [])
    fmt = kwargs.get('format', 'json')
    root = RootCmd(api=INTERNAL_API, resp_formatter_name=fmt)
    ret, _ = root([cmd] + params)
    if ret.status == STATUS_ERROR:
        raise ApiException(str(ret.value))
    return ret.value


@register(name="operator.show")
def operator_show(**kwargs):
    return operator_run('show', **kwargs)


@register(name="operator.set")
def operator_set(**kwargs):
    return operator_run('set', **kwargs)


@register(name="operator.clear")
def operator_clear(**kwargs):
    return operator_run('clear', **kwargs)



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Defines data types and models required specifically
 for L2VPN support.
 Represents data structures for VRF not VPN/global.
 (Inside VRF you have L2VPN Flow Specification prefixes
 and inside VPN you have L2VPN Flow Specification prefixes)
"""

import logging

from ryu.lib.packet.bgp import RF_L2VPN_FLOWSPEC
from ryu.lib.packet.bgp import FlowSpecL2VPNNLRI

from ryu.services.protocols.bgp.info_base.l2vpnfs import L2VPNFlowSpecPath
from ryu.services.protocols.bgp.info_base.vrffs import VRFFlowSpecDest
from ryu.services.protocols.bgp.info_base.vrffs import VRFFlowSpecPath
from ryu.services.protocols.bgp.info_base.vrffs import VRFFlowSpecTable

LOG = logging.getLogger('bgpspeaker.info_base.vrfl2vpnfs')


class L2vpnFlowSpecPath(VRFFlowSpecPath):
    """Represents a way of reaching an IP destination with
    a L2VPN Flow Specification.
    """
    ROUTE_FAMILY = RF_L2VPN_FLOWSPEC
    VPN_PATH_CLASS = L2VPNFlowSpecPath
    VPN_NLRI_CLASS = FlowSpecL2VPNNLRI


class L2vpnFlowSpecDest(VRFFlowSpecDest):
    ROUTE_FAMILY = RF_L2VPN_FLOWSPEC


class L2vpnFlowSpecTable(VRFFlowSpecTable):
    """Virtual Routing and Forwarding information base
    for L2VPN Flow Specification.
    """
    ROUTE_FAMILY = RF_L2VPN_FLOWSPEC
    VPN_ROUTE_FAMILY = RF_L2VPN_FLOWSPEC
    NLRI_CLASS = FlowSpecL2VPNNLRI
    VRF_PATH_CLASS = L2vpnFlowSpecPath
    VRF_DEST_CLASS = L2vpnFlowSpecDest



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 CLI application for SSH management.
"""

from copy import copy
import logging
import os.path
import sys

import paramiko

from ryu import version
from ryu.lib import hub
from ryu.services.protocols.bgp.base import Activity
from ryu.services.protocols.bgp.operator.command import Command
from ryu.services.protocols.bgp.operator.command import CommandsResponse
from ryu.services.protocols.bgp.operator.command import STATUS_OK
from ryu.services.protocols.bgp.operator.commands.root import RootCmd
from ryu.services.protocols.bgp.operator.internal_api import InternalApi

SSH_PORT = "ssh_port"
SSH_HOST = "ssh_host"
SSH_HOST_KEY = "ssh_host_key"
SSH_USERNAME = "ssh_username"
SSH_PASSWORD = "ssh_password"

DEFAULT_SSH_PORT = 4990
DEFAULT_SSH_HOST = "localhost"
DEFAULT_SSH_HOST_KEY = None
DEFAULT_SSH_USERNAME = "ryu"
DEFAULT_SSH_PASSWORD = "ryu"

CONF = {
    SSH_PORT: DEFAULT_SSH_PORT,
    SSH_HOST: DEFAULT_SSH_HOST,
    SSH_HOST_KEY: DEFAULT_SSH_HOST_KEY,
    SSH_USERNAME: DEFAULT_SSH_USERNAME,
    SSH_PASSWORD: DEFAULT_SSH_PASSWORD,
}

LOG = logging.getLogger('bgpspeaker.cli')


def find_ssh_server_key():
    if CONF[SSH_HOST_KEY]:
        return paramiko.RSAKey.from_private_key_file(CONF[SSH_HOST_KEY])
    elif os.path.exists("/etc/ssh_host_rsa_key"):
        # OSX
        return paramiko.RSAKey.from_private_key_file(
            "/etc/ssh_host_rsa_key")
    elif os.path.exists("/etc/ssh/ssh_host_rsa_key"):
        # Linux
        return paramiko.RSAKey.from_private_key_file(
            "/etc/ssh/ssh_host_rsa_key")
    else:
        return paramiko.RSAKey.generate(1024)


class SshServer(paramiko.ServerInterface):
    TERM = "ansi"
    PROMPT = "bgpd> "
    WELCOME = "\n\rHello, this is Ryu BGP speaker (version %s).\n\r" % version

    class HelpCmd(Command):
        help_msg = 'show this help'
        command = 'help'

        def action(self, params):
            return self.parent_cmd.question_mark()[0]

    class QuitCmd(Command):
        help_msg = 'exit this session'
        command = 'quit'

        def action(self, params):
            self.api.sshserver.end_session()
            return CommandsResponse(STATUS_OK, True)

    def __init__(self, sock, addr):
        super(SshServer, self).__init__()
        self.sock = sock
        self.addr = addr

        # tweak InternalApi and RootCmd for non-bgp related commands
        self.api = InternalApi(log_handler=logging.StreamHandler(sys.stderr))
        setattr(self.api, 'sshserver', self)
        self.root = RootCmd(self.api)
        self.root.subcommands['help'] = self.HelpCmd
        self.root.subcommands['quit'] = self.QuitCmd

        self.transport = paramiko.Transport(self.sock)
        self.transport.load_server_moduli()
        host_key = find_ssh_server_key()
        self.transport.add_server_key(host_key)
        self.transport.start_server(server=self)

        self.is_connected = True

        # For pylint
        self.buf = None
        self.chan = None
        self.curpos = None
        self.histindex = None
        self.history = None
        self.prompted = None
        self.promptlen = None

    def check_auth_none(self, username):
        return paramiko.AUTH_SUCCESSFUL

    def check_auth_password(self, username, password):
        if (username == CONF[SSH_USERNAME]
                and password == CONF[SSH_PASSWORD]):
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def check_channel_request(self, kind, chanid):
        if kind == 'session':
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_channel_shell_request(self, channel):
        hub.spawn(self._handle_shell_request)
        return True

    def check_channel_pty_request(self, channel, term, width, height,
                                  pixelwidth, pixelheight, modes):
        self.TERM = term
        return True

    def check_channel_window_change_request(self, channel, width, height,
                                            pixelwidth, pixelheight):
        return True

    @staticmethod
    def _is_echoable(c):
        return not (c < chr(0x20) or c == chr(0x7F))

    @staticmethod
    def _is_enter(c):
        return c == chr(0x0d)

    @staticmethod
    def _is_eof(c):
        return c == chr(0x03)

    @staticmethod
    def _is_esc(c):
        return c == chr(0x1b)

    @staticmethod
    def _is_hist(c):
        return c == chr(0x10) or c == chr(0x0e)

    @staticmethod
    def _is_del(c):
        return (c == chr(0x04) or c == chr(0x08) or c == chr(0x15)
                or c == chr(0x17) or c == chr(0x0c) or c == chr(0x7f))

    @staticmethod
    def _is_curmov(c):
        return c == chr(0x01) or c == chr(0x02) or c == chr(0x05) \
            or c == chr(0x06)

    @staticmethod
    def _is_cmpl(c):
        return c == chr(0x09)

    def _handle_csi_seq(self):
        c = self.chan.recv(1)
        c = c.decode()  # For Python3 compatibility
        if c == 'A':
            self._lookup_hist_up()
        elif c == 'B':
            self._lookup_hist_down()
        elif c == 'C':
            self._movcursor(self.curpos + 1)
        elif c == 'D':
            self._movcursor(self.curpos - 1)
        else:
            LOG.error("unknown CSI sequence. do nothing: %c", c)

    def _handle_esc_seq(self):
        c = self.chan.recv(1)
        c = c.decode()  # For Python3 compatibility
        if c == '[':
            self._handle_csi_seq()
        else:
            LOG.error("non CSI sequence. do nothing")

    def _send_csi_seq(self, cmd):
        self.chan.send('\x1b[' + cmd)

    def _movcursor(self, curpos):
        if self.prompted and curpos < len(self.PROMPT):
            self.curpos = len(self.PROMPT)
        elif self.prompted and curpos > (len(self.PROMPT) + len(self.buf)):
            self.curpos = len(self.PROMPT) + len(self.buf)
        else:
            self._send_csi_seq('%dG' % (curpos + 1))
            self.curpos = curpos

    def _clearscreen(self, prompt=None):
        if not prompt and self.prompted:
            prompt = self.PROMPT
        # clear screen
        self._send_csi_seq('2J')
        # move cursor to the top
        self._send_csi_seq('d')
        # redraw prompt and buf
        self._refreshline(prompt=prompt)

    def _clearline(self, prompt=None):
        if not prompt and self.prompted:
            prompt = self.PROMPT
        self.prompted = False
        self._movcursor(0)
        self._send_csi_seq('2K')
        if prompt:
            self.prompted = True
            self.chan.send(prompt)
            self._movcursor(len(prompt))
        self.buf = []

    def _refreshline(self, prompt=None):
        if not prompt and self.prompted:
            prompt = self.PROMPT
        buf = copy(self.buf)
        curpos = copy(self.curpos)
        self._clearline(prompt=prompt)
        self.chan.send(''.join(buf))
        self.buf = buf
        self.curpos = curpos
        self._movcursor(curpos)

    def _refreshnewline(self, prompt=None):
        if not prompt and self.prompted:
            prompt = self.PROMPT
        buf = copy(self.buf)
        curpos = copy(self.curpos)
        self._startnewline(prompt)
        self.chan.send(''.join(buf))
        self.buf = buf
        self.curpos = curpos
        self._movcursor(curpos)

    def _startnewline(self, prompt=None, buf=''):
        if not prompt and self.prompted:
            prompt = self.PROMPT
        if isinstance(buf, str):
            buf = list(buf)
        if self.chan:
            self.buf = buf
            if prompt:
                self.chan.send('\n\r' + prompt + ''.join(buf))
                self.curpos = len(prompt) + len(buf)
                self.prompted = True
            else:
                self.chan.send('\n\r' + ''.join(buf))
                self.curpos = len(buf)
                self.prompted = False

    def _lookup_hist_up(self):
        if len(self.history) == 0:
            return
        self.buf = self.history[self.histindex]
        self.curpos = self.promptlen + len(self.buf)
        self._refreshline()
        if self.histindex + 1 < len(self.history):
            self.histindex += 1

    def _lookup_hist_down(self):
        if self.histindex > 0:
            self.histindex -= 1
            self.buf = self.history[self.histindex]
            self.curpos = self.promptlen + len(self.buf)
            self._refreshline()
        else:
            self._clearline()

    def _do_cmpl(self, buf, is_exec=False):
        cmpleter = self.root
        is_spaced = buf[-1] == ' ' if len(buf) > 0 else False
        cmds = [tkn.strip() for tkn in ''.join(buf).split()]
        ret = []

        for i, cmd in enumerate(cmds):
            subcmds = cmpleter.subcommands
            matches = [x for x in subcmds.keys() if x.startswith(cmd)]

            if len(matches) == 1:
                cmpled_cmd = matches[0]
                cmpleter = subcmds[cmpled_cmd](self.api)

                if is_exec:
                    ret.append(cmpled_cmd)
                    continue

                if (i + 1) == len(cmds):
                    if is_spaced:
                        result, cmd = cmpleter('?')
                        result = result.value.replace('\n', '\n\r').rstrip()
                        self.prompted = False
                        buf = copy(buf)
                        self._startnewline(buf=result)
                        self.prompted = True
                        self._startnewline(buf=buf)
                    else:
                        self.buf = buf[:(-1 * len(cmd))] + \
                            list(cmpled_cmd + ' ')
                        self.curpos += len(cmpled_cmd) - len(cmd) + 1
                        self._refreshline()
            else:
                self.prompted = False
                buf = copy(self.buf)
                if len(matches) == 0:
                    if cmpleter.param_help_msg:
                        self.prompted = True
                        ret.append(cmd)
                        continue
                    else:
                        self._startnewline(buf='Error: Not implemented')
                else:
                    if (i + 1) < len(cmds):
                        self._startnewline(buf='Error: Ambiguous command')
                    else:
                        self._startnewline(buf=', '.join(matches))
                ret = []
                self.prompted = True
                if not is_exec:
                    self._startnewline(buf=buf)
                break

        return ret

    def _execute_cmd(self, cmds):
        result, _ = self.root(cmds)
        LOG.debug("result: %s", result)
        if cmds[0] == 'quit':
            self.is_connected = False
            return result.status
        self.prompted = False
        self._startnewline()
        output = result.value.replace('\n', '\n\r').rstrip()
        self.chan.send(output)
        self.prompted = True
        self._startnewline()
        return result.status

    def end_session(self):
        self._startnewline(prompt=False, buf='bye.\n\r')
        self.chan.close()

    def _handle_shell_request(self):
        LOG.info("session start")
        chan = self.transport.accept(20)
        if not chan:
            LOG.info("transport.accept timed out")
            return

        self.chan = chan
        self.buf = []
        self.curpos = 0
        self.history = []
        self.histindex = 0
        self.prompted = True
        self.chan.send(self.WELCOME)
        self._startnewline()

        while self.is_connected:
            c = self.chan.recv(1)
            c = c.decode()  # For Python3 compatibility

            if len(c) == 0:
                break

            LOG.debug("ord:%d, hex:0x%x", ord(c), ord(c))
            self.promptlen = len(self.PROMPT) if self.prompted else 0
            if c == '?':
                cmpleter = self.root
                cmds = [tkn.strip() for tkn in ''.join(self.buf).split()]

                for i, cmd in enumerate(cmds):
                    subcmds = cmpleter.subcommands
                    matches = [x for x in subcmds.keys() if x.startswith(cmd)]
                    if len(matches) == 1:
                        cmpled_cmd = matches[0]
                        cmpleter = subcmds[cmpled_cmd](self.api)

                result, cmd = cmpleter('?')
                result = result.value.replace('\n', '\n\r').rstrip()
                self.prompted = False
                buf = copy(self.buf)
                self._startnewline(buf=result)
                self.prompted = True
                self._startnewline(buf=buf)
            elif self._is_echoable(c):
                self.buf.insert(self.curpos - self.promptlen, c)
                self.curpos += 1
                self._refreshline()
            elif self._is_esc(c):
                self._handle_esc_seq()
            elif self._is_eof(c):
                self.end_session()
            elif self._is_curmov(c):
                # <C-a>
                if c == chr(0x01):
                    self._movcursor(self.promptlen)
                # <C-b>
                elif c == chr(0x02):
                    self._movcursor(self.curpos - 1)
                # <C-e>
                elif c == chr(0x05):
                    self._movcursor(self.promptlen + len(self.buf))
                # <C-f>
                elif c == chr(0x06):
                    self._movcursor(self.curpos + 1)
                else:
                    LOG.error("unknown cursor move cmd.")
                    continue
            elif self._is_hist(c):
                # <C-p>
                if c == chr(0x10):
                    self._lookup_hist_up()
                # <C-n>
                elif c == chr(0x0e):
                    self._lookup_hist_down()
            elif self._is_del(c):
                # <C-d>
                if c == chr(0x04):
                    if self.curpos < (self.promptlen + len(self.buf)):
                        self.buf.pop(self.curpos - self.promptlen)
                        self._refreshline()
                # <C-h> or delete
                elif c == chr(0x08) or c == chr(0x7f):
                    if self.curpos > self.promptlen:
                        self.buf.pop(self.curpos - self.promptlen - 1)
                        self.curpos -= 1
                        self._refreshline()
                # <C-u>
                elif c == chr(0x15):
                    self._clearline()
                # <C-w>
                elif c == chr(0x17):
                    pos = self.curpos - self.promptlen
                    i = pos
                    flag = False
                    for c in reversed(self.buf[:pos]):
                        if flag and c == ' ':
                            break
                        if c != ' ':
                            flag = True
                        i -= 1
                    del self.buf[i:pos]
                    self.curpos = self.promptlen + i
                    self._refreshline()
                # <C-l>
                elif c == chr(0x0c):
                    self._clearscreen()
            elif self._is_cmpl(c):
                self._do_cmpl(self.buf)
            elif self._is_enter(c):
                if len(''.join(self.buf).strip()) != 0:
                    # cmd line interpretation
                    cmds = self._do_cmpl(self.buf, is_exec=True)
                    if cmds:
                        self.history.insert(0, self.buf)
                        self.histindex = 0
                        self._execute_cmd(cmds)
                    else:
                        LOG.debug("no command is interpreted. "
                                  "just start a new line.")
                        self._startnewline()
                else:
                    LOG.debug("blank buf is detected. "
                              "just start a new line.")
                    self._startnewline()

            LOG.debug("curpos: %d, buf: %s, prompted: %s", self.curpos,
                      self.buf, self.prompted)

        LOG.info("session end")


def ssh_server_factory(sock, addr):
    SshServer(sock, addr)


class Cli(Activity):
    def __init__(self):
        super(Cli, self).__init__()

    def _run(self, *args, **kwargs):
        for k, v in kwargs.items():
            if k in CONF:
                CONF[k] = v

        listen_info = (CONF[SSH_HOST], CONF[SSH_PORT])
        LOG.info("starting ssh server at %s:%d" % listen_info)
        server = hub.StreamServer(listen_info, ssh_server_factory)
        server.serve_forever()

SSH_CLI_CONTROLLER = Cli()



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This sample application performs as VTEP for EVPN VXLAN and constructs
a Single Subnet per EVI corresponding to the VLAN Based service in [RFC7432].

.. NOTE::

    This app will invoke OVSDB request to the switches.
    Please set the manager address before calling the API of this app.

    ::

        $ sudo ovs-vsctl set-manager ptcp:6640
        $ sudo ovs-vsctl show
            ...(snip)
            Manager "ptcp:6640"
            ...(snip)


Usage Example
=============

Environment
-----------

This example supposes the following environment::

     Host A (172.17.0.1)                      Host B (172.17.0.2)
    +--------------------+                   +--------------------+
    |   Ryu1             | --- BGP(EVPN) --- |   Ryu2             |
    +--------------------+                   +--------------------+
            |                                       |
    +--------------------+                   +--------------------+
    |   s1 (OVS)         | ===== vxlan ===== |   s2 (OVS)         |
    +--------------------+                   +--------------------+
    (s1-eth1)    (s1-eth2)                   (s2-eth1)    (s2-eth2)
       |            |                           |            |
    +--------+  +--------+                   +--------+  +--------+
    | s1h1   |  | s1h2   |                   | s2h1   |  | s2h2   |
    +--------+  +--------+                   +--------+  +--------+

Configuration steps
-------------------

1. Creates a new BGPSpeaker instance on each host.

    On Host A::

        (Host A)$ curl -X POST -d '{
         "dpid": 1,
         "as_number": 65000,
         "router_id": "172.17.0.1"
         }' http://localhost:8080/vtep/speakers | python -m json.tool

    On Host B::

        (Host B)$ curl -X POST -d '{
         "dpid": 1,
         "as_number": 65000,
         "router_id": "172.17.0.2"
         }' http://localhost:8080/vtep/speakers | python -m json.tool

2. Registers the neighbor for the speakers on each host.

    On Host A::

        (Host A)$ curl -X POST -d '{
         "address": "172.17.0.2",
         "remote_as": 65000
         }' http://localhost:8080/vtep/neighbors |
         python -m json.tool

    On Host B::

        (Host B)$ curl -X POST -d '{
         "address": "172.17.0.1",
         "remote_as": 65000
         }' http://localhost:8080/vtep/neighbors |
         python -m json.tool

3. Defines a new VXLAN network(VNI=10) on the Host A/B.

    On Host A::

        (Host A)$ curl -X POST -d '{
         "vni": 10
         }' http://localhost:8080/vtep/networks | python -m json.tool

    On Host B::

        (Host B)$ curl -X POST -d '{
         "vni": 10
         }' http://localhost:8080/vtep/networks | python -m json.tool

4. Registers the clients to the VXLAN network.

    For "s1h1"(ip="10.0.0.11", mac="aa:bb:cc:00:00:11") on Host A::

        (Host A)$ curl -X POST -d '{
         "port": "s1-eth1",
         "mac": "aa:bb:cc:00:00:11",
         "ip": "10.0.0.11"
         } ' http://localhost:8080/vtep/networks/10/clients |
         python -m json.tool

    For "s2h1"(ip="10.0.0.21", mac="aa:bb:cc:00:00:21") on Host B::

        (Host B)$ curl -X POST -d '{
         "port": "s2-eth1",
         "mac": "aa:bb:cc:00:00:21",
         "ip": "10.0.0.21"
         } ' http://localhost:8080/vtep/networks/10/clients |
         python -m json.tool

Testing
-------

If BGP (EVPN) connection between Ryu1 and Ryu2 has been established,
pings between the client s1h1 and s2h1 should work.

::

    (s1h1)$ ping 10.0.0.21


Troubleshooting
---------------

If connectivity between s1h1 and s2h1 isn't working,
please check the followings.

1. Make sure that Host A and Host B have full network connectivity.

    ::

        (Host A)$ ping 172.17.0.2

2. Make sure that BGP(EVPN) connection has been established.

    ::

        (Host A)$ curl -X GET http://localhost:8080/vtep/neighbors |
         python -m json.tool

        ...
        {
            "172.17.0.2": {
                "EvpnNeighbor": {
                    "address": "172.17.0.2",
                    "remote_as": 65000,
                    "state": "up"  # "up" shows the connection established
                }
            }
        }

3. Make sure that BGP(EVPN) routes have been advertised.

    ::

        (Host A)$ curl -X GET http://localhost:8080/vtep/networks |
         python -m json.tool

         ...
        {
            "10": {
                "EvpnNetwork": {
                    "clients": {
                        "aa:bb:cc:00:00:11": {
                            "EvpnClient": {
                                "ip": "10.0.0.11",
                                "mac": "aa:bb:cc:00:00:11",
                                "next_hop": "172.17.0.1",
                                "port": 1
                            }
                        },
                        "aa:bb:cc:00:00:21": {  # route for "s2h1" on Host B
                            "EvpnClient": {
                                "ip": "10.0.0.21",
                                "mac": "aa:bb:cc:00:00:21",
                                "next_hop": "172.17.0.2",
                                "port": 3
                            }
                        }
                    },
                    "ethernet_tag_id": 0,
                    "route_dist": "65000:10",
                    "vni": 10
                }
            }
        }
"""

import json

from ryu.app.ofctl import api as ofctl_api
from ryu.app.wsgi import ControllerBase
from ryu.app.wsgi import Response
from ryu.app.wsgi import route
from ryu.app.wsgi import WSGIApplication
from ryu.base import app_manager
from ryu.exception import RyuException
from ryu.lib.ovs import bridge as ovs_bridge
from ryu.lib.packet import arp
from ryu.lib.packet import ether_types
from ryu.lib.packet.bgp import _RouteDistinguisher
from ryu.lib.packet.bgp import EvpnNLRI
from ryu.lib.stringify import StringifyMixin
from ryu.ofproto import ofproto_v1_3
from ryu.services.protocols.bgp.bgpspeaker import BGPSpeaker
from ryu.services.protocols.bgp.bgpspeaker import RF_L2_EVPN
from ryu.services.protocols.bgp.bgpspeaker import EVPN_MAC_IP_ADV_ROUTE
from ryu.services.protocols.bgp.bgpspeaker import EVPN_MULTICAST_ETAG_ROUTE
from ryu.services.protocols.bgp.info_base.evpn import EvpnPath


API_NAME = 'restvtep'

OVSDB_PORT = 6640  # The IANA registered port for OVSDB [RFC7047]

PRIORITY_D_PLANE = 1
PRIORITY_ARP_REPLAY = 2

TABLE_ID_INGRESS = 0
TABLE_ID_EGRESS = 1


# Utility functions

def to_int(i):
    return int(str(i), 0)


def to_str_list(l):
    str_list = []
    for s in l:
        str_list.append(str(s))
    return str_list


# Exception classes related to OpenFlow and OVSDB

class RestApiException(RyuException):

    def to_response(self, status):
        body = {
            "error": str(self),
            "status": status,
        }
        return Response(content_type='application/json',
                        body=json.dumps(body), status=status)


class DatapathNotFound(RestApiException):
    message = 'No such datapath: %(dpid)s'


class OFPortNotFound(RestApiException):
    message = 'No such OFPort: %(port_name)s'


# Exception classes related to BGP

class BGPSpeakerNotFound(RestApiException):
    message = 'BGPSpeaker could not be found'


class NeighborNotFound(RestApiException):
    message = 'No such neighbor: %(address)s'


class VniNotFound(RestApiException):
    message = 'No such VNI: %(vni)s'


class ClientNotFound(RestApiException):
    message = 'No such client: %(mac)s'


class ClientNotLocal(RestApiException):
    message = 'Specified client is not local: %(mac)s'


# Utility classes related to EVPN

class EvpnSpeaker(BGPSpeaker, StringifyMixin):
    _TYPE = {
        'ascii': [
            'router_id',
        ],
    }

    def __init__(self, dpid, as_number, router_id,
                 best_path_change_handler,
                 peer_down_handler, peer_up_handler,
                 neighbors=None):
        super(EvpnSpeaker, self).__init__(
            as_number=as_number,
            router_id=router_id,
            best_path_change_handler=best_path_change_handler,
            peer_down_handler=peer_down_handler,
            peer_up_handler=peer_up_handler,
            ssh_console=True)

        self.dpid = dpid
        self.as_number = as_number
        self.router_id = router_id
        self.neighbors = neighbors or {}


class EvpnNeighbor(StringifyMixin):
    _TYPE = {
        'ascii': [
            'address',
            'state',
        ],
    }

    def __init__(self, address, remote_as, state='down'):
        super(EvpnNeighbor, self).__init__()
        self.address = address
        self.remote_as = remote_as
        self.state = state


class EvpnNetwork(StringifyMixin):
    _TYPE = {
        'ascii': [
            'route_dist',
        ],
    }

    def __init__(self, vni, route_dist, ethernet_tag_id, clients=None):
        super(EvpnNetwork, self).__init__()
        self.vni = vni
        self.route_dist = route_dist
        self.ethernet_tag_id = ethernet_tag_id
        self.clients = clients or {}

    def get_clients(self, **kwargs):
        l = []
        for _, c in self.clients.items():
            for k, v in kwargs.items():
                if getattr(c, k) != v:
                    break
            else:
                l.append(c)
        return l


class EvpnClient(StringifyMixin):
    _TYPE = {
        'ascii': [
            'mac',
            'ip',
            'next_hop'
        ],
    }

    def __init__(self, port, mac, ip, next_hop):
        super(EvpnClient, self).__init__()
        self.port = port
        self.mac = mac
        self.ip = ip
        self.next_hop = next_hop


class RestVtep(app_manager.RyuApp):
    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]
    _CONTEXTS = {'wsgi': WSGIApplication}

    def __init__(self, *args, **kwargs):
        super(RestVtep, self).__init__(*args, **kwargs)
        wsgi = kwargs['wsgi']
        wsgi.register(RestVtepController, {RestVtep.__name__: self})

        # EvpnSpeaker instance instantiated later
        self.speaker = None

        # OVSBridge instance instantiated later
        self.ovs = None

        # Dictionary for retrieving the EvpnNetwork instance by VNI
        # self.networks = {
        #     <vni>: <instance 'EvpnNetwork'>,
        #     ...
        # }
        self.networks = {}

    # Utility methods related to OpenFlow

    def _get_datapath(self, dpid):
        return ofctl_api.get_datapath(self, dpid)

    @staticmethod
    def _add_flow(datapath, priority, match, instructions,
                  table_id=TABLE_ID_INGRESS):
        parser = datapath.ofproto_parser

        mod = parser.OFPFlowMod(
            datapath=datapath,
            table_id=table_id,
            priority=priority,
            match=match,
            instructions=instructions)

        datapath.send_msg(mod)

    @staticmethod
    def _del_flow(datapath, priority, match, table_id=TABLE_ID_INGRESS):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        mod = parser.OFPFlowMod(
            datapath=datapath,
            table_id=table_id,
            command=ofproto.OFPFC_DELETE,
            priority=priority,
            out_port=ofproto.OFPP_ANY,
            out_group=ofproto.OFPG_ANY,
            match=match)

        datapath.send_msg(mod)

    def _add_network_ingress_flow(self, datapath, tag, in_port, eth_src=None):
        parser = datapath.ofproto_parser

        if eth_src is None:
            match = parser.OFPMatch(in_port=in_port)
        else:
            match = parser.OFPMatch(in_port=in_port, eth_src=eth_src)
        instructions = [
            parser.OFPInstructionWriteMetadata(
                metadata=tag, metadata_mask=parser.UINT64_MAX),
            parser.OFPInstructionGotoTable(1)]

        self._add_flow(datapath, PRIORITY_D_PLANE, match, instructions)

    def _del_network_ingress_flow(self, datapath, in_port, eth_src=None):
        parser = datapath.ofproto_parser

        if eth_src is None:
            match = parser.OFPMatch(in_port=in_port)
        else:
            match = parser.OFPMatch(in_port=in_port, eth_src=eth_src)

        self._del_flow(datapath, PRIORITY_D_PLANE, match)

    def _add_arp_reply_flow(self, datapath, tag, arp_tpa, arp_tha):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        match = parser.OFPMatch(
            metadata=(tag, parser.UINT64_MAX),
            eth_type=ether_types.ETH_TYPE_ARP,
            arp_op=arp.ARP_REQUEST,
            arp_tpa=arp_tpa)

        actions = [
            parser.NXActionRegMove(
                src_field="eth_src", dst_field="eth_dst", n_bits=48),
            parser.OFPActionSetField(eth_src=arp_tha),
            parser.OFPActionSetField(arp_op=arp.ARP_REPLY),
            parser.NXActionRegMove(
                src_field="arp_sha", dst_field="arp_tha", n_bits=48),
            parser.NXActionRegMove(
                src_field="arp_spa", dst_field="arp_tpa", n_bits=32),
            parser.OFPActionSetField(arp_sha=arp_tha),
            parser.OFPActionSetField(arp_spa=arp_tpa),
            parser.OFPActionOutput(ofproto.OFPP_IN_PORT)]
        instructions = [
            parser.OFPInstructionActions(
                ofproto.OFPIT_APPLY_ACTIONS, actions)]

        self._add_flow(datapath, PRIORITY_ARP_REPLAY, match, instructions,
                       table_id=TABLE_ID_EGRESS)

    def _del_arp_reply_flow(self, datapath, tag, arp_tpa):
        parser = datapath.ofproto_parser

        match = parser.OFPMatch(
            metadata=(tag, parser.UINT64_MAX),
            eth_type=ether_types.ETH_TYPE_ARP,
            arp_op=arp.ARP_REQUEST,
            arp_tpa=arp_tpa)

        self._del_flow(datapath, PRIORITY_ARP_REPLAY, match,
                       table_id=TABLE_ID_EGRESS)

    def _add_l2_switching_flow(self, datapath, tag, eth_dst, out_port):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        match = parser.OFPMatch(metadata=(tag, parser.UINT64_MAX),
                                eth_dst=eth_dst)
        actions = [parser.OFPActionOutput(out_port)]
        instructions = [
            parser.OFPInstructionActions(
                ofproto.OFPIT_APPLY_ACTIONS, actions)]

        self._add_flow(datapath, PRIORITY_D_PLANE, match, instructions,
                       table_id=TABLE_ID_EGRESS)

    def _del_l2_switching_flow(self, datapath, tag, eth_dst):
        parser = datapath.ofproto_parser

        match = parser.OFPMatch(metadata=(tag, parser.UINT64_MAX),
                                eth_dst=eth_dst)

        self._del_flow(datapath, PRIORITY_D_PLANE, match,
                       table_id=TABLE_ID_EGRESS)

    def _del_network_egress_flow(self, datapath, tag):
        parser = datapath.ofproto_parser

        match = parser.OFPMatch(metadata=(tag, parser.UINT64_MAX))

        self._del_flow(datapath, PRIORITY_D_PLANE, match,
                       table_id=TABLE_ID_EGRESS)

    # Utility methods related to OVSDB

    def _get_ovs_bridge(self, dpid):
        datapath = self._get_datapath(dpid)
        if datapath is None:
            self.logger.debug('No such datapath: %s', dpid)
            return None

        ovsdb_addr = 'tcp:%s:%d' % (datapath.address[0], OVSDB_PORT)
        if (self.ovs is not None
                and self.ovs.datapath_id == dpid
                and self.ovs.vsctl.remote == ovsdb_addr):
            return self.ovs

        try:
            self.ovs = ovs_bridge.OVSBridge(
                CONF=self.CONF,
                datapath_id=datapath.id,
                ovsdb_addr=ovsdb_addr)
            self.ovs.init()
        except Exception as e:
            self.logger.exception('Cannot initiate OVSDB connection: %s', e)
            return None

        return self.ovs

    def _get_ofport(self, dpid, port_name):
        ovs = self._get_ovs_bridge(dpid)
        if ovs is None:
            return None

        try:
            return ovs.get_ofport(port_name)
        except Exception as e:
            self.logger.debug('Cannot get port number for %s: %s',
                              port_name, e)
            return None

    def _get_vxlan_port(self, dpid, remote_ip, key):
        # Searches VXLAN port named 'vxlan_<remote_ip>_<key>'
        return self._get_ofport(dpid, 'vxlan_%s_%s' % (remote_ip, key))

    def _add_vxlan_port(self, dpid, remote_ip, key):
        # If VXLAN port already exists, returns OFPort number
        vxlan_port = self._get_vxlan_port(dpid, remote_ip, key)
        if vxlan_port is not None:
            return vxlan_port

        ovs = self._get_ovs_bridge(dpid)
        if ovs is None:
            return None

        # Adds VXLAN port named 'vxlan_<remote_ip>_<key>'
        ovs.add_vxlan_port(
            name='vxlan_%s_%s' % (remote_ip, key),
            remote_ip=remote_ip,
            key=key)

        # Returns VXLAN port number
        return self._get_vxlan_port(dpid, remote_ip, key)

    def _del_vxlan_port(self, dpid, remote_ip, key):
        ovs = self._get_ovs_bridge(dpid)
        if ovs is None:
            return None

        # If VXLAN port does not exist, returns None
        vxlan_port = self._get_vxlan_port(dpid, remote_ip, key)
        if vxlan_port is None:
            return None

        # Adds VXLAN port named 'vxlan_<remote_ip>_<key>'
        ovs.del_port('vxlan_%s_%s' % (remote_ip, key))

        # Returns deleted VXLAN port number
        return vxlan_port

    # Event handlers for BGP

    def _evpn_mac_ip_adv_route_handler(self, ev):
        network = self.networks.get(ev.path.nlri.vni, None)
        if network is None:
            self.logger.debug('No such VNI registered: %s', ev.path.nlri)
            return

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            self.logger.debug('No such datapath: %s', self.speaker.dpid)
            return

        vxlan_port = self._add_vxlan_port(
            dpid=self.speaker.dpid,
            remote_ip=ev.nexthop,
            key=ev.path.nlri.vni)
        if vxlan_port is None:
            self.logger.debug('Cannot create a new VXLAN port: %s',
                              'vxlan_%s_%s' % (ev.nexthop, ev.path.nlri.vni))
            return

        self._add_l2_switching_flow(
            datapath=datapath,
            tag=network.vni,
            eth_dst=ev.path.nlri.mac_addr,
            out_port=vxlan_port)

        self._add_arp_reply_flow(
            datapath=datapath,
            tag=network.vni,
            arp_tpa=ev.path.nlri.ip_addr,
            arp_tha=ev.path.nlri.mac_addr)

        network.clients[ev.path.nlri.mac_addr] = EvpnClient(
            port=vxlan_port,
            mac=ev.path.nlri.mac_addr,
            ip=ev.path.nlri.ip_addr,
            next_hop=ev.nexthop)

    def _evpn_incl_mcast_etag_route_handler(self, ev):
        # Note: For the VLAN Based service, we use RT(=RD) assigned
        # field as vid.
        vni = _RouteDistinguisher.from_str(ev.path.nlri.route_dist).assigned

        network = self.networks.get(vni, None)
        if network is None:
            self.logger.debug('No such VNI registered: %s', vni)
            return

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            self.logger.debug('No such datapath: %s', self.speaker.dpid)
            return

        vxlan_port = self._add_vxlan_port(
            dpid=self.speaker.dpid,
            remote_ip=ev.nexthop,
            key=vni)
        if vxlan_port is None:
            self.logger.debug('Cannot create a new VXLAN port: %s',
                              'vxlan_%s_%s' % (ev.nexthop, vni))
            return

        self._add_network_ingress_flow(
            datapath=datapath,
            tag=vni,
            in_port=vxlan_port)

    def _evpn_route_handler(self, ev):
        if ev.path.nlri.type == EvpnNLRI.MAC_IP_ADVERTISEMENT:
            self._evpn_mac_ip_adv_route_handler(ev)
        elif ev.path.nlri.type == EvpnNLRI.INCLUSIVE_MULTICAST_ETHERNET_TAG:
            self._evpn_incl_mcast_etag_route_handler(ev)

    def _evpn_withdraw_mac_ip_adv_route_handler(self, ev):
        network = self.networks.get(ev.path.nlri.vni, None)
        if network is None:
            self.logger.debug('No such VNI registered: %s', ev.path.nlri)
            return

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            self.logger.debug('No such datapath: %s', self.speaker.dpid)
            return

        client = network.clients.get(ev.path.nlri.mac_addr, None)
        if client is None:
            self.logger.debug('No such client: %s', ev.path.nlri.mac_addr)
            return

        self._del_l2_switching_flow(
            datapath=datapath,
            tag=network.vni,
            eth_dst=ev.path.nlri.mac_addr)

        self._del_arp_reply_flow(
            datapath=datapath,
            tag=network.vni,
            arp_tpa=ev.path.nlri.ip_addr)

        network.clients.pop(ev.path.nlri.mac_addr)

    def _evpn_withdraw_incl_mcast_etag_route_handler(self, ev):
        # Note: For the VLAN Based service, we use RT(=RD) assigned
        # field as vid.
        vni = _RouteDistinguisher.from_str(ev.path.nlri.route_dist).assigned

        network = self.networks.get(vni, None)
        if network is None:
            self.logger.debug('No such VNI registered: %s', vni)
            return

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            self.logger.debug('No such datapath: %s', self.speaker.dpid)
            return

        vxlan_port = self._get_vxlan_port(
            dpid=self.speaker.dpid,
            remote_ip=ev.nexthop,
            key=vni)
        if vxlan_port is None:
            self.logger.debug('No such VXLAN port: %s',
                              'vxlan_%s_%s' % (ev.nexthop, vni))
            return

        self._del_network_ingress_flow(
            datapath=datapath,
            in_port=vxlan_port)

        vxlan_port = self._del_vxlan_port(
            dpid=self.speaker.dpid,
            remote_ip=ev.nexthop,
            key=vni)
        if vxlan_port is None:
            self.logger.debug('Cannot delete VXLAN port: %s',
                              'vxlan_%s_%s' % (ev.nexthop, vni))
            return

    def _evpn_withdraw_route_handler(self, ev):
        if ev.path.nlri.type == EvpnNLRI.MAC_IP_ADVERTISEMENT:
            self._evpn_withdraw_mac_ip_adv_route_handler(ev)
        elif ev.path.nlri.type == EvpnNLRI.INCLUSIVE_MULTICAST_ETHERNET_TAG:
            self._evpn_withdraw_incl_mcast_etag_route_handler(ev)

    def _best_path_change_handler(self, ev):
        if not isinstance(ev.path, EvpnPath):
            # Ignores non-EVPN routes
            return
        elif ev.nexthop == self.speaker.router_id:
            # Ignore local connected routes
            return
        elif ev.is_withdraw:
            self._evpn_withdraw_route_handler(ev)
        else:
            self._evpn_route_handler(ev)

    def _peer_down_handler(self, remote_ip, remote_as):
        neighbor = self.speaker.neighbors.get(remote_ip, None)
        if neighbor is None:
            self.logger.debug('No such neighbor: remote_ip=%s, remote_as=%s',
                              remote_ip, remote_as)
            return

        neighbor.state = 'down'

    def _peer_up_handler(self, remote_ip, remote_as):
        neighbor = self.speaker.neighbors.get(remote_ip, None)
        if neighbor is None:
            self.logger.debug('No such neighbor: remote_ip=%s, remote_as=%s',
                              remote_ip, remote_as)
            return

        neighbor.state = 'up'

    # API methods for REST controller

    def add_speaker(self, dpid, as_number, router_id):
        # Check if the datapath for the specified dpid exist or not
        datapath = self._get_datapath(dpid)
        if datapath is None:
            raise DatapathNotFound(dpid=dpid)

        self.speaker = EvpnSpeaker(
            dpid=dpid,
            as_number=as_number,
            router_id=router_id,
            best_path_change_handler=self._best_path_change_handler,
            peer_down_handler=self._peer_down_handler,
            peer_up_handler=self._peer_up_handler)

        return {self.speaker.router_id: self.speaker.to_jsondict()}

    def get_speaker(self):
        if self.speaker is None:
            return BGPSpeakerNotFound()

        return {self.speaker.router_id: self.speaker.to_jsondict()}

    def del_speaker(self):
        if self.speaker is None:
            return BGPSpeakerNotFound()

        for vni in list(self.networks.keys()):
            self.del_network(vni=vni)

        for address in list(self.speaker.neighbors.keys()):
            self.del_neighbor(address=address)

        self.speaker.shutdown()
        speaker = self.speaker
        self.speaker = None

        return {speaker.router_id: speaker.to_jsondict()}

    def add_neighbor(self, address, remote_as):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        self.speaker.neighbor_add(
            address=address,
            remote_as=remote_as,
            enable_evpn=True)

        neighbor = EvpnNeighbor(
            address=address,
            remote_as=remote_as)
        self.speaker.neighbors[address] = neighbor

        return {address: neighbor.to_jsondict()}

    def get_neighbors(self, address=None):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        if address is not None:
            neighbor = self.speaker.neighbors.get(address, None)
            if neighbor is None:
                raise NeighborNotFound(address=address)
            return {address: neighbor.to_jsondict()}

        neighbors = {}
        for address, neighbor in self.speaker.neighbors.items():
            neighbors[address] = neighbor.to_jsondict()

        return neighbors

    def del_neighbor(self, address):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        neighbor = self.speaker.neighbors.get(address, None)
        if neighbor is None:
            raise NeighborNotFound(address=address)

        for network in self.networks.values():
            for mac, client in list(network.clients.items()):
                if client.next_hop == address:
                    network.clients.pop(mac)

        self.speaker.neighbor_del(address=address)

        neighbor = self.speaker.neighbors.pop(address)

        return {address: neighbor.to_jsondict()}

    def add_network(self, vni):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        # Constructs type 0 RD with as_number and vni
        route_dist = "%s:%d" % (self.speaker.as_number, vni)

        self.speaker.vrf_add(
            route_dist=route_dist,
            import_rts=[route_dist],
            export_rts=[route_dist],
            route_family=RF_L2_EVPN)

        # Note: For the VLAN Based service, ethernet_tag_id
        # must be set to zero.
        self.speaker.evpn_prefix_add(
            route_type=EVPN_MULTICAST_ETAG_ROUTE,
            route_dist=route_dist,
            ethernet_tag_id=vni,
            ip_addr=self.speaker.router_id,
            next_hop=self.speaker.router_id)

        network = EvpnNetwork(
            vni=vni,
            route_dist=route_dist,
            ethernet_tag_id=0)
        self.networks[vni] = network

        return {vni: network.to_jsondict()}

    def get_networks(self, vni=None):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        if vni is not None:
            network = self.networks.get(vni, None)
            if network is None:
                raise VniNotFound(vni=vni)
            return {vni: network.to_jsondict()}

        networks = {}
        for vni, network in self.networks.items():
            networks[vni] = network.to_jsondict()

        return networks

    def del_network(self, vni):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            raise DatapathNotFound(dpid=self.speaker.dpid)

        network = self.networks.get(vni, None)
        if network is None:
            raise VniNotFound(vni=vni)

        for client in network.get_clients(next_hop=self.speaker.router_id):
            self.del_client(
                vni=vni,
                mac=client.mac)

        self._del_network_egress_flow(
            datapath=datapath,
            tag=vni)

        for address in self.speaker.neighbors:
            self._del_vxlan_port(
                dpid=self.speaker.dpid,
                remote_ip=address,
                key=vni)

        self.speaker.evpn_prefix_del(
            route_type=EVPN_MULTICAST_ETAG_ROUTE,
            route_dist=network.route_dist,
            ethernet_tag_id=vni,
            ip_addr=self.speaker.router_id)

        self.speaker.vrf_del(route_dist=network.route_dist)

        network = self.networks.pop(vni)

        return {vni: network.to_jsondict()}

    def add_client(self, vni, port, mac, ip):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            raise DatapathNotFound(dpid=self.speaker.dpid)

        network = self.networks.get(vni, None)
        if network is None:
            raise VniNotFound(vni=vni)

        port = self._get_ofport(self.speaker.dpid, port)
        if port is None:
            try:
                port = to_int(port)
            except ValueError:
                raise OFPortNotFound(port_name=port)

        self._add_network_ingress_flow(
            datapath=datapath,
            tag=network.vni,
            in_port=port,
            eth_src=mac)

        self._add_l2_switching_flow(
            datapath=datapath,
            tag=network.vni,
            eth_dst=mac,
            out_port=port)

        # Note: For the VLAN Based service, ethernet_tag_id
        # must be set to zero.
        self.speaker.evpn_prefix_add(
            route_type=EVPN_MAC_IP_ADV_ROUTE,
            route_dist=network.route_dist,
            esi=0,
            ethernet_tag_id=0,
            mac_addr=mac,
            ip_addr=ip,
            vni=vni,
            next_hop=self.speaker.router_id,
            tunnel_type='vxlan')

        # Stores local client info
        client = EvpnClient(
            port=port,
            mac=mac,
            ip=ip,
            next_hop=self.speaker.router_id)
        network.clients[mac] = client

        return {vni: client.to_jsondict()}

    def del_client(self, vni, mac):
        if self.speaker is None:
            raise BGPSpeakerNotFound()

        datapath = self._get_datapath(self.speaker.dpid)
        if datapath is None:
            raise DatapathNotFound(dpid=self.speaker.dpid)

        network = self.networks.get(vni, None)
        if network is None:
            raise VniNotFound(vni=vni)

        client = network.clients.get(mac, None)
        if client is None:
            raise ClientNotFound(mac=mac)
        elif client.next_hop != self.speaker.router_id:
            raise ClientNotLocal(mac=mac)

        self._del_network_ingress_flow(
            datapath=datapath,
            in_port=client.port,
            eth_src=mac)

        self._del_l2_switching_flow(
            datapath=datapath,
            tag=network.vni,
            eth_dst=mac)

        # Note: For the VLAN Based service, ethernet_tag_id
        # must be set to zero.
        self.speaker.evpn_prefix_del(
            route_type=EVPN_MAC_IP_ADV_ROUTE,
            route_dist=network.route_dist,
            esi=0,
            ethernet_tag_id=0,
            mac_addr=mac,
            ip_addr=client.ip)

        client = network.clients.pop(mac)

        return {vni: client.to_jsondict()}


def post_method(keywords):
    def _wrapper(method):
        def __wrapper(self, req, **kwargs):
            try:
                try:
                    body = req.json if req.body else {}
                except ValueError:
                    raise ValueError('Invalid syntax %s', req.body)
                kwargs.update(body)
                for key, converter in keywords.items():
                    value = kwargs.get(key, None)
                    if value is None:
                        raise ValueError('%s not specified' % key)
                    kwargs[key] = converter(value)
            except ValueError as e:
                return Response(content_type='application/json',
                                body={"error": str(e)}, status=400)
            try:
                return method(self, **kwargs)
            except Exception as e:
                status = 500
                body = {
                    "error": str(e),
                    "status": status,
                }
                return Response(content_type='application/json',
                                body=json.dumps(body), status=status)
        __wrapper.__doc__ = method.__doc__
        return __wrapper
    return _wrapper


def get_method(keywords=None):
    keywords = keywords or {}

    def _wrapper(method):
        def __wrapper(self, _, **kwargs):
            try:
                for key, converter in keywords.items():
                    value = kwargs.get(key, None)
                    if value is None:
                        continue
                    kwargs[key] = converter(value)
            except ValueError as e:
                return Response(content_type='application/json',
                                body={"error": str(e)}, status=400)
            try:
                return method(self, **kwargs)
            except Exception as e:
                status = 500
                body = {
                    "error": str(e),
                    "status": status,
                }
                return Response(content_type='application/json',
                                body=json.dumps(body), status=status)
        __wrapper.__doc__ = method.__doc__
        return __wrapper
    return _wrapper


delete_method = get_method


class RestVtepController(ControllerBase):

    def __init__(self, req, link, data, **config):
        super(RestVtepController, self).__init__(req, link, data, **config)
        self.vtep_app = data[RestVtep.__name__]
        self.logger = self.vtep_app.logger

    @route(API_NAME, '/vtep/speakers', methods=['POST'])
    @post_method(
        keywords={
            "dpid": to_int,
            "as_number": to_int,
            "router_id": str,
        })
    def add_speaker(self, **kwargs):
        """
        Creates a new BGPSpeaker instance.

        Usage:

            ======= ================
            Method  URI
            ======= ================
            POST    /vtep/speakers
            ======= ================

        Request parameters:

            ========== ============================================
            Attribute  Description
            ========== ============================================
            dpid       ID of Datapath binding to speaker. (e.g. 1)
            as_number  AS number. (e.g. 65000)
            router_id  Router ID. (e.g. "172.17.0.1")
            ========== ============================================

        Example::

            $ curl -X POST -d '{
             "dpid": 1,
             "as_number": 65000,
             "router_id": "172.17.0.1"
             }' http://localhost:8080/vtep/speakers | python -m json.tool

        ::

            {
                "172.17.0.1": {
                    "EvpnSpeaker": {
                        "as_number": 65000,
                        "dpid": 1,
                        "neighbors": {},
                        "router_id": "172.17.0.1"
                    }
                }
            }
        """
        try:
            body = self.vtep_app.add_speaker(**kwargs)
        except DatapathNotFound as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/speakers', methods=['GET'])
    @get_method()
    def get_speakers(self, **kwargs):
        """
        Gets the info of BGPSpeaker instance.

        Usage:

            ======= ================
            Method  URI
            ======= ================
            GET     /vtep/speakers
            ======= ================

        Example::

            $ curl -X GET http://localhost:8080/vtep/speakers |
             python -m json.tool

        ::

            {
                "172.17.0.1": {
                    "EvpnSpeaker": {
                        "as_number": 65000,
                        "dpid": 1,
                        "neighbors": {
                            "172.17.0.2": {
                                "EvpnNeighbor": {
                                    "address": "172.17.0.2",
                                    "remote_as": 65000,
                                    "state": "up"
                                }
                            }
                        },
                        "router_id": "172.17.0.1"
                    }
                }
            }
        """
        try:
            body = self.vtep_app.get_speaker()
        except BGPSpeakerNotFound as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/speakers', methods=['DELETE'])
    @delete_method()
    def del_speaker(self, **kwargs):
        """
        Shutdowns BGPSpeaker instance.

        Usage:

            ======= ================
            Method  URI
            ======= ================
            DELETE  /vtep/speakers
            ======= ================

        Example::

            $ curl -X DELETE http://localhost:8080/vtep/speakers |
             python -m json.tool

        ::

            {
                "172.17.0.1": {
                    "EvpnSpeaker": {
                        "as_number": 65000,
                        "dpid": 1,
                        "neighbors": {},
                        "router_id": "172.17.0.1"
                    }
                }
            }
        """
        try:
            body = self.vtep_app.del_speaker()
        except BGPSpeakerNotFound as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/neighbors', methods=['POST'])
    @post_method(
        keywords={
            "address": str,
            "remote_as": to_int,
        })
    def add_neighbor(self, **kwargs):
        """
        Registers a new neighbor to the speaker.

        Usage:

            ======= ========================
            Method  URI
            ======= ========================
            POST    /vtep/neighbors
            ======= ========================

        Request parameters:

            ========== ================================================
            Attribute  Description
            ========== ================================================
            address    IP address of neighbor. (e.g. "172.17.0.2")
            remote_as  AS number of neighbor. (e.g. 65000)
            ========== ================================================

        Example::

            $ curl -X POST -d '{
             "address": "172.17.0.2",
             "remote_as": 65000
             }' http://localhost:8080/vtep/neighbors |
             python -m json.tool

        ::

            {
                "172.17.0.2": {
                    "EvpnNeighbor": {
                        "address": "172.17.0.2",
                        "remote_as": 65000,
                        "state": "down"
                    }
                }
            }
        """
        try:
            body = self.vtep_app.add_neighbor(**kwargs)
        except BGPSpeakerNotFound as e:
            return e.to_response(status=400)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    def _get_neighbors(self, **kwargs):
        try:
            body = self.vtep_app.get_neighbors(**kwargs)
        except (BGPSpeakerNotFound, NeighborNotFound) as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/neighbors', methods=['GET'])
    @get_method()
    def get_neighbors(self, **kwargs):
        """
        Gets a list of all neighbors.

        Usage:

            ======= ========================
            Method  URI
            ======= ========================
            GET     /vtep/neighbors
            ======= ========================

        Example::

            $ curl -X GET http://localhost:8080/vtep/neighbors |
             python -m json.tool

        ::

            {
                "172.17.0.2": {
                    "EvpnNeighbor": {
                        "address": "172.17.0.2",
                        "remote_as": 65000,
                        "state": "up"
                    }
                }
            }
        """
        return self._get_neighbors(**kwargs)

    @route(API_NAME, '/vtep/neighbors/{address}', methods=['GET'])
    @get_method(
        keywords={
            "address": str,
        })
    def get_neighbor(self, **kwargs):
        """
        Gets the neighbor for the specified address.

        Usage:

            ======= ==================================
            Method  URI
            ======= ==================================
            GET     /vtep/neighbors/{address}
            ======= ==================================

        Request parameters:

            ========== ================================================
            Attribute  Description
            ========== ================================================
            address    IP address of neighbor. (e.g. "172.17.0.2")
            ========== ================================================

        Example::

            $ curl -X GET http://localhost:8080/vtep/neighbors/172.17.0.2 |
             python -m json.tool

        ::

            {
                "172.17.0.2": {
                    "EvpnNeighbor": {
                        "address": "172.17.0.2",
                        "remote_as": 65000,
                        "state": "up"
                    }
                }
            }
        """
        return self._get_neighbors(**kwargs)

    @route(API_NAME, '/vtep/neighbors/{address}', methods=['DELETE'])
    @delete_method(
        keywords={
            "address": str,
        })
    def del_neighbor(self, **kwargs):
        """
        Unregister the specified neighbor from the speaker.

        Usage:

            ======= ==================================
            Method  URI
            ======= ==================================
            DELETE  /vtep/speaker/neighbors/{address}
            ======= ==================================

        Request parameters:

            ========== ================================================
            Attribute  Description
            ========== ================================================
            address    IP address of neighbor. (e.g. "172.17.0.2")
            ========== ================================================

        Example::

            $ curl -X DELETE http://localhost:8080/vtep/speaker/neighbors/172.17.0.2 |
             python -m json.tool

        ::

            {
                "172.17.0.2": {
                    "EvpnNeighbor": {
                        "address": "172.17.0.2",
                        "remote_as": 65000,
                        "state": "up"
                    }
                }
            }
        """
        try:
            body = self.vtep_app.del_neighbor(**kwargs)
        except (BGPSpeakerNotFound, NeighborNotFound) as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/networks', methods=['POST'])
    @post_method(
        keywords={
            "vni": to_int,
        })
    def add_network(self, **kwargs):
        """
        Defines a new network.

        Usage:

            ======= ===============
            Method  URI
            ======= ===============
            POST    /vtep/networks
            ======= ===============

        Request parameters:

            ================ ========================================
            Attribute        Description
            ================ ========================================
            vni              Virtual Network Identifier. (e.g. 10)
            ================ ========================================

        Example::

            $ curl -X POST -d '{
             "vni": 10
             }' http://localhost:8080/vtep/networks | python -m json.tool

        ::

            {
                "10": {
                    "EvpnNetwork": {
                        "clients": {},
                        "ethernet_tag_id": 0,
                        "route_dist": "65000:10",
                        "vni": 10
                    }
                }
            }
        """
        try:
            body = self.vtep_app.add_network(**kwargs)
        except BGPSpeakerNotFound as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    def _get_networks(self, **kwargs):
        try:
            body = self.vtep_app.get_networks(**kwargs)
        except (BGPSpeakerNotFound, VniNotFound) as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/networks', methods=['GET'])
    @get_method()
    def get_networks(self, **kwargs):
        """
        Gets a list of all networks.

        Usage:

            ======= ===============
            Method  URI
            ======= ===============
            GET     /vtep/networks
            ======= ===============

        Example::

            $ curl -X GET http://localhost:8080/vtep/networks |
             python -m json.tool

        ::

            {
                "10": {
                    "EvpnNetwork": {
                        "clients": {
                            "aa:bb:cc:dd:ee:ff": {
                                "EvpnClient": {
                                    "ip": "10.0.0.1",
                                    "mac": "aa:bb:cc:dd:ee:ff",
                                    "next_hop": "172.17.0.1",
                                    "port": 1
                                }
                            }
                        },
                        "ethernet_tag_id": 0,
                        "route_dist": "65000:10",
                        "vni": 10
                    }
                }
            }
        """
        return self._get_networks(**kwargs)

    @route(API_NAME, '/vtep/networks/{vni}', methods=['GET'])
    @get_method(
        keywords={
            "vni": to_int,
        })
    def get_network(self, **kwargs):
        """
        Gets the network for the specified VNI.

        Usage:

            ======= =====================
            Method  URI
            ======= =====================
            GET     /vtep/networks/{vni}
            ======= =====================

        Request parameters:

            ================ ========================================
            Attribute        Description
            ================ ========================================
            vni              Virtual Network Identifier. (e.g. 10)
            ================ ========================================

        Example::

            $ curl -X GET http://localhost:8080/vtep/networks/10 |
             python -m json.tool

        ::

            {
                "10": {
                    "EvpnNetwork": {
                        "clients": {
                            "aa:bb:cc:dd:ee:ff": {
                                "EvpnClient": {
                                    "ip": "10.0.0.1",
                                    "mac": "aa:bb:cc:dd:ee:ff",
                                    "next_hop": "172.17.0.1",
                                    "port": 1
                                }
                            }
                        },
                        "ethernet_tag_id": 0,
                        "route_dist": "65000:10",
                        "vni": 10
                    }
                }
            }
        """
        return self._get_networks(**kwargs)

    @route(API_NAME, '/vtep/networks/{vni}', methods=['DELETE'])
    @delete_method(
        keywords={
            "vni": to_int,
        })
    def del_network(self, **kwargs):
        """
        Deletes the network for the specified VNI.

        Usage:

            ======= =====================
            Method  URI
            ======= =====================
            DELETE  /vtep/networks/{vni}
            ======= =====================

        Request parameters:

            ================ ========================================
            Attribute        Description
            ================ ========================================
            vni              Virtual Network Identifier. (e.g. 10)
            ================ ========================================

        Example::

            $ curl -X DELETE http://localhost:8080/vtep/networks/10 |
             python -m json.tool

        ::

            {
                "10": {
                    "EvpnNetwork": {
                        "ethernet_tag_id": 10,
                        "clients": [
                            {
                                "EvpnClient": {
                                    "ip": "10.0.0.11",
                                    "mac": "e2:b1:0c:ba:42:ed",
                                    "port": 1
                                }
                            }
                        ],
                        "route_dist": "65000:100",
                        "vni": 10
                    }
                }
            }
        """
        try:
            body = self.vtep_app.del_network(**kwargs)
        except (BGPSpeakerNotFound, DatapathNotFound, VniNotFound) as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/networks/{vni}/clients', methods=['POST'])
    @post_method(
        keywords={
            "vni": to_int,
            "port": str,
            "mac": str,
            "ip": str,
        })
    def add_client(self, **kwargs):
        """
        Registers a new client to the specified network.

        Usage:

            ======= =============================
            Method  URI
            ======= =============================
            POST    /vtep/networks/{vni}/clients
            ======= =============================

        Request parameters:

            =========== ===============================================
            Attribute   Description
            =========== ===============================================
            vni         Virtual Network Identifier. (e.g. 10)
            port        Port number to connect client.
                        For convenience, port name can be specified
                        and automatically translated to port number.
                        (e.g. "s1-eth1" or 1)
            mac         Client MAC address to register.
                        (e.g. "aa:bb:cc:dd:ee:ff")
            ip          Client IP address. (e.g. "10.0.0.1")
            =========== ===============================================

        Example::

            $ curl -X POST -d '{
             "port": "s1-eth1",
             "mac": "aa:bb:cc:dd:ee:ff",
             "ip": "10.0.0.1"
             }' http://localhost:8080/vtep/networks/10/clients |
             python -m json.tool

        ::

            {
                "10": {
                    "EvpnClient": {
                        "ip": "10.0.0.1",
                        "mac": "aa:bb:cc:dd:ee:ff",
                        "next_hop": "172.17.0.1",
                        "port": 1
                    }
                }
            }
        """
        try:
            body = self.vtep_app.add_client(**kwargs)
        except (BGPSpeakerNotFound, DatapathNotFound,
                VniNotFound, OFPortNotFound) as e:
            return e.to_response(status=404)

        return Response(content_type='application/json',
                        body=json.dumps(body))

    @route(API_NAME, '/vtep/networks/{vni}/clients/{mac}', methods=['DELETE'])
    @delete_method(
        keywords={
            "vni": to_int,
            "mac": str,
        })
    def del_client(self, **kwargs):
        """
        Registers a new client to the specified network.

        Usage:

            ======= ===================================
            Method  URI
            ======= ===================================
            DELETE  /vtep/networks/{vni}/clients/{mac}
            ======= ===================================

        Request parameters:

            =========== ===============================================
            Attribute   Description
            =========== ===============================================
            vni         Virtual Network Identifier. (e.g. 10)
            mac         Client MAC address to register.
            =========== ===============================================

        Example::

            $ curl -X DELETE http://localhost:8080/vtep/networks/10/clients/aa:bb:cc:dd:ee:ff |
             python -m json.tool

        ::

            {
                "10": {
                    "EvpnClient": {
                        "ip": "10.0.0.1",
                        "mac": "aa:bb:cc:dd:ee:ff",
                        "next_hop": "172.17.0.1",
                        "port": 1
                    }
                }
            }
        """
        try:
            body = self.vtep_app.del_client(**kwargs)
        except (BGPSpeakerNotFound, DatapathNotFound,
                VniNotFound, ClientNotFound, ClientNotLocal) as e:
            return Response(body=str(e), status=500)

        return Response(content_type='application/json',
                        body=json.dumps(body))



# Copyright (C) 2011 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ryu.base import app_manager
from ryu.controller import ofp_event
from ryu.controller.handler import CONFIG_DISPATCHER, MAIN_DISPATCHER
from ryu.controller.handler import set_ev_cls
from ryu.ofproto import ofproto_v1_3
from ryu.lib.packet import packet
from ryu.lib.packet import ethernet
from ryu.lib.packet import ether_types


class SimpleSwitch13(app_manager.RyuApp):
    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]

    def __init__(self, *args, **kwargs):
        super(SimpleSwitch13, self).__init__(*args, **kwargs)
        self.mac_to_port = {}

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)
    def switch_features_handler(self, ev):
        datapath = ev.msg.datapath
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        # install table-miss flow entry
        #
        # We specify NO BUFFER to max_len of the output action due to
        # OVS bug. At this moment, if we specify a lesser number, e.g.,
        # 128, OVS will send Packet-In with invalid buffer_id and
        # truncated packet data. In that case, we cannot output packets
        # correctly.  The bug has been fixed in OVS v2.1.0.
        match = parser.OFPMatch()
        actions = [parser.OFPActionOutput(ofproto.OFPP_CONTROLLER,
                                          ofproto.OFPCML_NO_BUFFER)]
        self.add_flow(datapath, 0, match, actions)

    def add_flow(self, datapath, priority, match, actions, buffer_id=None):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        inst = [parser.OFPInstructionActions(ofproto.OFPIT_APPLY_ACTIONS,
                                             actions)]
        if buffer_id:
            mod = parser.OFPFlowMod(datapath=datapath, buffer_id=buffer_id,
                                    priority=priority, match=match,
                                    instructions=inst)
        else:
            mod = parser.OFPFlowMod(datapath=datapath, priority=priority,
                                    match=match, instructions=inst)
        datapath.send_msg(mod)

    @set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)
    def _packet_in_handler(self, ev):
        # If you hit this you might want to increase
        # the "miss_send_length" of your switch
        if ev.msg.msg_len < ev.msg.total_len:
            self.logger.debug("packet truncated: only %s of %s bytes",
                              ev.msg.msg_len, ev.msg.total_len)
        msg = ev.msg
        datapath = msg.datapath
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        in_port = msg.match['in_port']

        pkt = packet.Packet(msg.data)
        eth = pkt.get_protocols(ethernet.ethernet)[0]

        if eth.ethertype == ether_types.ETH_TYPE_LLDP:
            # ignore lldp packet
            return
        dst = eth.dst
        src = eth.src

        dpid = datapath.id
        self.mac_to_port.setdefault(dpid, {})

        self.logger.info("packet in %s %s %s %s", dpid, src, dst, in_port)

        # learn a mac address to avoid FLOOD next time.
        self.mac_to_port[dpid][src] = in_port

        if dst in self.mac_to_port[dpid]:
            out_port = self.mac_to_port[dpid][dst]
        else:
            out_port = ofproto.OFPP_FLOOD

        actions = [parser.OFPActionOutput(out_port)]

        # install a flow to avoid packet_in next time
        if out_port != ofproto.OFPP_FLOOD:
            match = parser.OFPMatch(in_port=in_port, eth_dst=dst)
            # verify if we have a valid buffer_id, if yes avoid to send both
            # flow_mod & packet_out
            if msg.buffer_id != ofproto.OFP_NO_BUFFER:
                self.add_flow(datapath, 1, match, actions, msg.buffer_id)
                return
            else:
                self.add_flow(datapath, 1, match, actions)
        data = None
        if msg.buffer_id == ofproto.OFP_NO_BUFFER:
            data = msg.data

        out = parser.OFPPacketOut(datapath=datapath, buffer_id=msg.buffer_id,
                                  in_port=in_port, actions=actions, data=data)
        datapath.send_msg(out)



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils.spawn import find_executable
import logging
import subprocess
import unittest

from nose.tools import eq_
from nose.tools import ok_

from ryu.lib.hub import sleep
from ryu.lib.ovs import vsctl


LOG = logging.getLogger(__name__)

DOCKER_IMAGE_MININET = 'osrg/ryu-book'

OVSDB_MANAGER_ADDR = 'ptcp:6640'
OVSDB_SWITCH_ADDR = 'tcp:%s:6640'


def setUpModule():
    if not find_executable('docker'):
        raise unittest.SkipTest(
            'Docker is not available. Test in %s will be skipped.' % __name__)


def _run(command):
    popen = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    popen.wait()
    result = popen.stdout.read().decode('utf-8')

    if result:
        return [str(r.strip('\r')) for r in result.split('\n')]
    else:
        return []


class TestVSCtl(unittest.TestCase):
    """
    Test cases for ryu.lib.ovs.vsctl.VSCtl
    """
    container_mn = None  # Container ID of Mininet
    container_mn_ip = None  # IP of Mininet container

    vsctl = None  # instance of vsctl.VSCtl

    @classmethod
    def _docker_exec(cls, container, command):
        return _run('docker exec -t %s %s' % (container, command))

    @classmethod
    def _docker_exec_mn(cls, command):
        return cls._docker_exec(cls.container_mn, command)

    @classmethod
    def _docker_run(cls, image):
        return _run('docker run --privileged -t -d %s' % image)[0]

    @classmethod
    def _docker_stop(cls, container):
        return _run('docker stop %s' % container)[0]

    @classmethod
    def _docker_rm(cls, container):
        return _run('docker rm %s' % container)[0]

    @classmethod
    def _docker_inspect_ip_addr(cls, container):
        return _run(
            'docker inspect --format="{{.NetworkSettings.IPAddress}}" %s' %
            container)[0].strip('"')

    @classmethod
    def _set_up_mn_container(cls):
        cls.container_mn = cls._docker_run(DOCKER_IMAGE_MININET)
        cls.container_mn_ip = cls._docker_inspect_ip_addr(cls.container_mn)

        # Note: Wait for loading the OVS kernel module.
        # If the OVS kernel module is loaded at first time, it might take
        # a few seconds.
        sleep(5)

        cls._docker_exec_mn(
            'ovs-vsctl set-manager %s' % OVSDB_MANAGER_ADDR)

    @classmethod
    def _set_up_vsctl(cls):
        cls.vsctl = vsctl.VSCtl(OVSDB_SWITCH_ADDR % cls.container_mn_ip)

    @classmethod
    def setUpClass(cls):
        cls._set_up_mn_container()
        cls._set_up_vsctl()

    @classmethod
    def _tear_down_mn_container(cls):
        cls._docker_exec_mn('mn --clean')
        cls._docker_stop(cls.container_mn)
        cls._docker_rm(cls.container_mn)

    @classmethod
    def tearDownClass(cls):
        cls._tear_down_mn_container()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def _run_commands(self, commands):
        self.vsctl.run_command(commands, timeout_sec=1)

    # 00: Open vSwitch commands

    def test_00_01_init(self):
        command = vsctl.VSCtlCommand('init')
        self._run_commands([command])

        ok_(command.result is None)

    def test_00_02_show(self):
        command = vsctl.VSCtlCommand('show')
        self._run_commands([command])

        ok_(command.result is not None)

    # 01: Bridge commands

    def test_01_01_add_br_bridge(self):
        bridge = 's1'
        command = vsctl.VSCtlCommand('add-br', (bridge,))
        self._run_commands([command])

        result = self._docker_exec_mn('ovs-vsctl list-br')
        ok_(bridge in result)

    def test_01_02_add_br_parent_vlan(self):
        bridge = 'sub-s1-100'
        parent = 's1'
        vlan = '100'
        command = vsctl.VSCtlCommand('add-br', (bridge, parent, vlan))
        self._run_commands([command])

        result = self._docker_exec_mn('ovs-vsctl list-br')
        ok_(bridge in result)
        result = self._docker_exec_mn(
            'ovs-vsctl br-to-parent %s' % bridge)
        eq_(parent, result[0])
        result = self._docker_exec_mn(
            'ovs-vsctl br-to-vlan %s' % bridge)
        eq_(vlan, result[0])

    def test_01_03_del_br(self):
        bridge = 's1'
        child = 'sub-s1-100'

        command = vsctl.VSCtlCommand('del-br', (bridge,))
        self._run_commands([command])

        result = self._docker_exec_mn('ovs-vsctl list-br')
        ok_(bridge not in result)
        ok_(child not in result)

    def test_01_04_list_br(self):
        bridge = 's1'
        child = 'sub-s1-100'
        vlan = '100'
        self._docker_exec_mn('ovs-vsctl add-br %s' % bridge)
        self._docker_exec_mn(
            'ovs-vsctl add-br %s %s %s' % (child, bridge, vlan))

        command = vsctl.VSCtlCommand('list-br')
        self._run_commands([command])

        ok_(bridge in command.result)
        ok_(child in command.result)

    def test_01_05_br_exists(self):
        bridge = 's1'

        command = vsctl.VSCtlCommand('br-exists', (bridge, ))
        self._run_commands([command])

        eq_(True, command.result)

    def test_01_06_br_to_vlan(self):
        bridge = 's1'

        command = vsctl.VSCtlCommand('br-to-vlan', (bridge, ))
        self._run_commands([command])

        eq_(0, command.result)

    def test_01_06_br_to_vlan_fake_bridge(self):
        bridge = 'sub-s1-100'

        command = vsctl.VSCtlCommand('br-to-vlan', (bridge, ))
        self._run_commands([command])

        eq_(100, command.result)

    def test_01_07_br_to_parent(self):
        bridge = 's1'
        parent = bridge

        command = vsctl.VSCtlCommand('br-to-parent', (bridge, ))
        self._run_commands([command])

        # result = <ryu.lib.ovs.vsctl.VSCtlBridge object>
        eq_(parent, command.result.name)

    def test_01_07_br_to_parent_fake_bridge(self):
        bridge = 'sub-s1-100'
        parent = 's1'

        command = vsctl.VSCtlCommand('br-to-parent', (bridge, ))
        self._run_commands([command])

        # result = <ryu.lib.ovs.vsctl.VSCtlBridge object>
        eq_(parent, command.result.name)

    def test_01_08_br_set_external_id_add(self):
        bridge = 's1'
        key = 'ext_id_key'
        value = 'ext_id_value'

        command = vsctl.VSCtlCommand(
            'br-set-external-id', (bridge, key, value))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl br-get-external-id %s %s' % (bridge, key))
        eq_(value, result[0])

    def test_01_09_br_get_external_id_value(self):
        bridge = 's1'
        key = 'ext_id_key'
        value = 'ext_id_value'

        command = vsctl.VSCtlCommand(
            'br-get-external-id', (bridge, key))
        self._run_commands([command])

        eq_(value, command.result)

    def test_01_10_br_get_external_id_dict(self):
        bridge = 's1'
        key = 'ext_id_key'
        value = 'ext_id_value'

        command = vsctl.VSCtlCommand(
            'br-get-external-id', (bridge,))
        self._run_commands([command])

        eq_({key: value}, command.result)

    def test_01_11_br_set_external_id_clear(self):
        bridge = 's1'
        key = 'ext_id_key'

        command = vsctl.VSCtlCommand(
            'br-set-external-id', (bridge, key))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl br-get-external-id %s %s' % (bridge, key))
        eq_([], result)

        # Clean up
        self._docker_exec_mn('mn --clean')

    # 02: Port commands

    def test_02_01_list_ports(self):
        bridge = 's1'
        interface_1 = 's1-eth1'
        interface_2 = 's1-eth2'

        self._docker_exec_mn(
            'ip link add %s type dummy' % interface_1)
        self._docker_exec_mn(
            'ip link add %s type dummy' % interface_2)
        self._docker_exec_mn(
            'ovs-vsctl add-br %(bridge)s'
            ' -- add-port %(bridge)s %(interface_1)s'
            ' -- add-port %(bridge)s %(interface_2)s' % locals())

        command = vsctl.VSCtlCommand('list-ports', (bridge,))
        self._run_commands([command])

        ok_(interface_1 in command.result)
        ok_(interface_2 in command.result)

    def test_02_02_add_port(self):
        bridge = 's1'
        interface_1 = 's1-eth1'
        self._docker_exec_mn(
            'ovs-vsctl del-port %s %s' % (bridge, interface_1))

        command = vsctl.VSCtlCommand('add-port', (bridge, interface_1))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl port-to-br %s' % interface_1)
        eq_(bridge, result[0])

    def test_02_03_add_bond(self):
        bridge = 's1'
        interface_1 = 's1-eth1'
        interface_2 = 's1-eth2'
        port = 's1-bond1'
        interface_list = [interface_1, interface_2]
        self._docker_exec_mn('ovs-vsctl del-br %s' % bridge)
        self._docker_exec_mn('ovs-vsctl add-br %s' % bridge)

        command = vsctl.VSCtlCommand(
            'add-bond', (bridge, port, interface_list))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl port-to-br %s' % port)
        eq_(bridge, result[0])

    def test_02_04_del_port(self):
        bridge = 's1'
        port = 's1-bond1'

        command = vsctl.VSCtlCommand('del-port', (bridge, port))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl list-ports %s' % bridge)
        eq_([], result)

    def test_02_05_port_to_br(self):
        bridge = 's1'
        port_1 = 's1-eth1'
        port_2 = 's1-eth2'
        self._docker_exec_mn('ovs-vsctl del-br %s' % bridge)
        self._docker_exec_mn(
            'ovs-vsctl add-br %(bridge)s'
            ' -- add-port %(bridge)s %(port_1)s'
            ' -- add-port %(bridge)s %(port_2)s' % locals())

        command = vsctl.VSCtlCommand('port-to-br', (port_1,))
        self._run_commands([command])

        eq_(bridge, command.result)

        # Clean up
        self._docker_exec_mn('mn --clean')

    # 03: Interface commands

    def test_03_01_list_ifaces(self):
        bridge = 's1'
        interface_1 = 's1-eth1'
        interface_2 = 's1-eth2'

        self._docker_exec_mn(
            'ip link add %s type dummy' % interface_1)
        self._docker_exec_mn(
            'ip link add %s type dummy' % interface_2)
        self._docker_exec_mn(
            'ovs-vsctl add-br %(bridge)s'
            ' -- add-port %(bridge)s %(interface_1)s'
            ' -- add-port %(bridge)s %(interface_2)s' % locals())

        command = vsctl.VSCtlCommand('list-ifaces', (bridge,))
        self._run_commands([command])

        ok_(interface_1 in command.result)
        ok_(interface_2 in command.result)

    def test_03_02_ifaces_to_br(self):
        bridge = 's1'
        interface_1 = 's1-eth1'

        command = vsctl.VSCtlCommand('iface-to-br', (interface_1,))
        self._run_commands([command])

        eq_(bridge, command.result)

        # Clean up
        self._docker_exec_mn('mn --clean')

    # 04: Controller commands

    def test_04_01_get_controller(self):
        bridge = 's1'
        controller = 'tcp:127.0.0.1:6653'
        self._docker_exec_mn(
            'ovs-vsctl add-br %(bridge)s'
            ' -- set-controller %(bridge)s %(controller)s' % locals())

        command = vsctl.VSCtlCommand('get-controller', (bridge,))
        self._run_commands([command])

        eq_(1, len(command.result))
        eq_(controller, command.result[0])

    def test_04_02_del_controller(self):
        bridge = 's1'

        command = vsctl.VSCtlCommand('del-controller', (bridge,))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get-controller %s' % bridge)
        eq_([], result)

    def test_04_03_set_controller(self):
        bridge = 's1'
        controller = 'tcp:127.0.0.1:6653'

        command = vsctl.VSCtlCommand('set-controller', (bridge, controller))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get-controller %s' % bridge)
        eq_(controller, result[0])

    def test_04_04_get_fail_mode(self):
        bridge = 's1'
        fai_mode = 'secure'
        self._docker_exec_mn(
            'ovs-vsctl set-fail-mode %s %s' % (bridge, fai_mode))

        command = vsctl.VSCtlCommand('get-fail-mode', (bridge,))
        self._run_commands([command])

        eq_(fai_mode, command.result)

    def test_04_05_del_fail_mode(self):
        bridge = 's1'

        command = vsctl.VSCtlCommand('del-fail-mode', (bridge,))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get-fail-mode %s' % bridge)
        eq_([], result)

    def test_04_06_set_fail_mode(self):
        bridge = 's1'
        fail_mode = 'secure'

        command = vsctl.VSCtlCommand('set-fail-mode', (bridge, fail_mode))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get-fail-mode %s' % bridge)
        eq_(fail_mode, result[0])

        # Clean up
        self._docker_exec_mn('mn --clean')

    # 05: Manager commands (not implemented yet)
    # 06: SSL commands (not implemented yet)
    # 07: Switch commands (not implemented yet)

    # 08: Database commands

    def test_08_01_list(self):
        table = 'Bridge'
        bridge = 's1'
        interface_1 = 's1-eth1'
        interface_2 = 's1-eth2'
        fail_mode = 'secure'
        protocols = 'OpenFlow10,OpenFlow13'
        datapath_id = '1111111111111111'

        self._docker_exec_mn(
            'ip link add %s type dummy' % interface_1)
        self._docker_exec_mn(
            'ip link add %s type dummy' % interface_2)
        self._docker_exec_mn(
            'ovs-vsctl add-br %(bridge)s'
            ' -- add-port %(bridge)s %(interface_1)s'
            ' -- add-port %(bridge)s %(interface_2)s' % locals())
        self._docker_exec_mn(
            'ovs-vsctl set %(table)s %(bridge)s '
            'fail_mode=%(fail_mode)s '
            'protocols=%(protocols)s '
            'other_config:datapath-id=%(datapath_id)s' % locals())

        command = vsctl.VSCtlCommand('list', (table,))
        self._run_commands([command])

        eq_(1, len(command.result))
        # command.result[0] = <ryu.lib.ovs.vsctl.VSCtlBridge object>
        eq_(bridge, command.result[0].name)

    def test_08_02_find(self):
        table = 'Bridge'
        bridge = 's1'

        command = vsctl.VSCtlCommand('find', (table, 'name=%s' % bridge))
        self._run_commands([command])

        eq_(1, len(command.result))
        # command.result[0] = <ovs.db.idl.Row object object> for Bridge
        eq_(bridge, command.result[0].name)

    def test_08_02_find_complex(self):
        table = 'Bridge'
        bridge = 's1'
        fail_mode = 'secure'
        protocols = 'OpenFlow10,OpenFlow13'
        datapath_id = '1111111111111111'

        command = vsctl.VSCtlCommand(
            'find', (table, 'fail_mode=%s' % fail_mode,
                     'protocols=%s' % protocols,
                     'other_config:datapath-id=%s' % datapath_id))
        self._run_commands([command])

        eq_(1, len(command.result))
        # command.result[0] = <ovs.db.idl.Row object object> for Bridge
        eq_(bridge, command.result[0].name)

    def test_08_03_get_01_value(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'fail_mode'
        value = 'secure'

        command = vsctl.VSCtlCommand('get', (table, bridge, column))
        self._run_commands([command])

        # command.result[0] is a list of return values
        eq_(value, command.result[0][0])

    def test_08_03_get_02_set(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'protocols'
        value = 'OpenFlow10,OpenFlow13'.split(',')

        command = vsctl.VSCtlCommand('get', (table, bridge, column))
        self._run_commands([command])

        # command.result[0] is a list
        eq_(value, command.result[0])

    def test_08_03_get_03_map(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'other_config'
        key = 'datapath-id'
        datapath_id = '1111111111111111'
        value = {key: datapath_id}

        command = vsctl.VSCtlCommand('get', (table, bridge, column))
        self._run_commands([command])

        # command.result[0] is a dict
        eq_(value, command.result[0])

    def test_08_03_get_04_map_value(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'other_config'
        key = 'datapath-id'
        datapath_id = '1111111111111111'
        value = datapath_id

        command = vsctl.VSCtlCommand(
            'get', (table, bridge, '%s:%s' % (column, key)))
        self._run_commands([command])

        # command.result[0] is a dict
        eq_(value, command.result[0])

    def test_08_04_set_01_value(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'fail_mode'
        value = 'standalone'

        command = vsctl.VSCtlCommand(
            'set', (table, bridge, '%s=%s' % (column, value)))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        eq_(value, result[0])

    def test_08_04_set_02_set(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'protocols'
        value = 'OpenFlow10,OpenFlow12,OpenFlow13'

        command = vsctl.VSCtlCommand(
            'set', (table, bridge, '%s=%s' % (column, value)))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '["OpenFlow10", "OpenFlow12", "OpenFlow13"]'
        eq_(expected_value, result[0])

    def test_08_04_set_03_map(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'other_config'
        key = 'datapath-id'
        value = '0000000000000001'

        command = vsctl.VSCtlCommand(
            'set', (table, bridge, '%s:%s=%s' % (column, key, value)))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s:%s' % (table, bridge, column, key))
        expected_value = '"0000000000000001"'
        eq_(expected_value, result[0])

    def test_08_05_add_01_value(self):
        table = 'Port'
        bridge = 's1'
        column = 'tag'
        value = '100'

        command = vsctl.VSCtlCommand(
            'add', (table, bridge, column, value))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        eq_(value, result[0])

    def test_08_05_add_02_set(self):
        table = 'Port'
        bridge = 's1'
        column = 'trunks'
        value = '100,200'

        command = vsctl.VSCtlCommand(
            'add', (table, bridge, column, value))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '[100, 200]'
        eq_(expected_value, result[0])

    def test_08_05_add_03_map(self):
        table = 'Bridge'
        bridge = 's1'
        column = 'other_config'
        key = 'datapath-id'
        value = '0000000000000011'

        command = vsctl.VSCtlCommand(
            'add', (table, bridge, column, '%s=%s' % (key, value)))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s:%s' % (table, bridge, column, key))
        expected_value = '"0000000000000011"'
        eq_(expected_value, result[0])

    def test_08_06_remove_01_value(self):
        table = 'Port'
        bridge = 's1'
        column = 'tag'
        value = '100'
        self._docker_exec_mn(
            'ovs-vsctl set %s %s %s=%s' % (table, bridge, column, value))

        command = vsctl.VSCtlCommand(
            'remove', (table, bridge, column, value))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '[]'
        eq_(expected_value, result[0])

    def test_08_06_remove_02_set(self):
        table = 'Port'
        bridge = 's1'
        column = 'trunks'
        init_value = '100,200,300'
        value = '100,200'
        self._docker_exec_mn(
            'ovs-vsctl set %s %s %s=%s' % (table, bridge, column, init_value))

        command = vsctl.VSCtlCommand(
            'remove', (table, bridge, column, value))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '[300]'
        eq_(expected_value, result[0])

    def test_08_06_remove_03_map(self):
        table = 'Port'
        bridge = 's1'
        column = 'other_config'
        key = 'priority-tag'
        value = 'true'
        self._docker_exec_mn(
            'ovs-vsctl set %s %s %s:%s=%s' %
            (table, bridge, column, key, value))

        command = vsctl.VSCtlCommand(
            'remove', (table, bridge, column, '%s=%s' % (key, value)))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '{}'
        eq_(expected_value, result[0])

    def test_08_07_clear_01_value(self):
        table = 'Port'
        bridge = 's1'
        column = 'tag'
        value = '100'
        self._docker_exec_mn(
            'ovs-vsctl set %s %s %s=%s' % (table, bridge, column, value))

        command = vsctl.VSCtlCommand(
            'clear', (table, bridge, column))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '[]'
        eq_(expected_value, result[0])

    def test_08_07_clear_02_set(self):
        table = 'Port'
        bridge = 's1'
        column = 'trunks'
        value = '100,200'
        self._docker_exec_mn(
            'ovs-vsctl set %s %s %s=%s' % (table, bridge, column, value))

        command = vsctl.VSCtlCommand(
            'clear', (table, bridge, column))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '[]'
        eq_(expected_value, result[0])

    def test_08_07_clear_03_map(self):
        table = 'Port'
        bridge = 's1'
        column = 'other_config'
        key = 'priority-tag'
        value = 'true'
        self._docker_exec_mn(
            'ovs-vsctl set %s %s %s:%s=%s' %
            (table, bridge, column, key, value))

        command = vsctl.VSCtlCommand(
            'clear', (table, bridge, column, '%s=%s' % (key, value)))
        self._run_commands([command])

        result = self._docker_exec_mn(
            'ovs-vsctl get %s %s %s' % (table, bridge, column))
        expected_value = '{}'
        eq_(expected_value, result[0])

        # Clean up
        self._docker_exec_mn('mn --clean')



# Copyright (C) 2016 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import bz2
import io
import logging
import os
import sys
import unittest

try:
    import mock  # Python 2
except ImportError:
    from unittest import mock  # Python 3

from nose.tools import eq_
from nose.tools import ok_

from ryu.lib import addrconv
from ryu.lib import mrtlib
from ryu.lib.packet import bgp
from ryu.lib.packet import ospf
from ryu.utils import binary_str


LOG = logging.getLogger(__name__)

MRT_DATA_DIR = os.path.join(
    os.path.dirname(sys.modules[__name__].__file__), '../../packet_data/mrt/')


class TestMrtlib(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.
    """

    def test_reader(self):
        files = [
            'rib.20161101.0000_pick.bz2',
            'updates.20161101.0000.bz2',
        ]

        for f in files:
            # print('\n*** testing mrtlib.Reader with %s ...' % f)
            counter = 0
            input_file = os.path.join(MRT_DATA_DIR, f)
            for record in mrtlib.Reader(bz2.BZ2File(input_file, 'rb')):
                # print('* No.%d\n%s' % (counter, record))
                ok_(not isinstance(record, mrtlib.UnknownMrtRecord))
                counter += 1

    def test_writer(self):
        files = [
            'rib.20161101.0000_pick.bz2',
            'updates.20161101.0000.bz2',
        ]

        for f in files:
            # print('\n*** testing mrtlib.Writer with %s ...' % f)
            input_file = os.path.join(MRT_DATA_DIR, f)
            input_buf = bz2.BZ2File(input_file, 'rb').read()
            input_records = list(mrtlib.Reader(bz2.BZ2File(input_file, 'rb')))

            counter = 0
            f = io.BytesIO()
            mrt_writer = mrtlib.Writer(f)
            for record in input_records:
                # print('* No.%d\n%s' % (counter, record))
                mrt_writer.write(record)
                counter += 1

            output_buf = f.getvalue()

            eq_(binary_str(input_buf), binary_str(output_buf))

            mrt_writer.close()

            eq_(True, mrt_writer._f.closed)


class TestMrtlibMrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.MrtRecord.
    """

    def test_init_without_type_subtype(self):
        type_ = mrtlib.MrtRecord.TYPE_TABLE_DUMP
        subtype = mrtlib.TableDumpMrtRecord.SUBTYPE_AFI_IPv4

        message = mrtlib.TableDumpAfiIPv4MrtMessage(
            view_num=1,
            seq_num=2,
            prefix='192.168.1.0',
            prefix_len=24,
            status=1,
            originated_time=0,
            peer_ip='10.0.0.1',
            peer_as=65000,
            bgp_attributes=[],
        )
        record = mrtlib.TableDumpMrtRecord(message)

        eq_(type_, record.type)
        eq_(subtype, record.subtype)

    def test_parse_pre_with_type_et(self):
        buf = (
            b'\x00\x00\x00\x00'  # timestamp
            b'\x00\x11\x00\x00'  # type=TYPE_BGP4MP_ET(17), subtype
            b'\x00\x00\x00\xaa'  # length
        )

        required_len = mrtlib.MrtRecord.parse_pre(buf)

        eq_(0xaa + mrtlib.ExtendedTimestampMrtRecord.HEADER_SIZE,
            required_len)


# Note: MrtCommonRecord is tested in TestMrtlibMrtRecord.
# class TestMrtlibMrtCommonRecord(unittest.TestCase):


class TestMrtlibExtendedTimestampMrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.ExtendedTimestampMrtRecord.
    """

    def test_parse_extended_header(self):
        body = b'test'
        buf = (
            b'\x11\x11\x11\x11'  # ms_timestamp
            + body
        )

        (headers,
         rest) = mrtlib.ExtendedTimestampMrtRecord.parse_extended_header(buf)

        ok_(isinstance(headers, list))
        eq_(1, len(headers))
        eq_(0x11111111, headers[0])
        eq_(body, rest)

    def test_serialize(self):
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x22\x22\x33\x33'  # type, subtype
            b'\x00\x00\x00\x04'  # length=len(body)
            b'\x44\x44\x44\x44'  # ms_timestamp
            + body
        )

        message_mock = mock.MagicMock(spec=mrtlib.MrtMessage)
        message_mock.serialize.return_value = body

        record = mrtlib.ExtendedTimestampMrtRecord(
            message=message_mock,
            timestamp=0x11111111,
            type_=0x2222, subtype=0x3333,
            ms_timestamp=0x44444444,
            length=0x00000004,
        )

        output = record.serialize()

        eq_(buf, output)


class TestMrtlibUnknownMrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.UnknownMrtRecord.
    """

    def test_parse(self):
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x22\x22\x33\x33'  # type, subtype
            b'\x00\x00\x00\x04'  # length=len(body)
            + body
        )

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(0x2222, record.type)
        eq_(0x3333, record.subtype)
        eq_(0x00000004, record.length)
        eq_(body, record.message.buf)
        eq_(b'', rest)

    def test_serialize(self):
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x22\x22\x33\x33'  # type, subtype
            b'\x00\x00\x00\x04'  # length=len(body)
            + body
        )

        message = mrtlib.UnknownMrtMessage(buf=body)
        record = mrtlib.UnknownMrtRecord(
            message=message,
            timestamp=0x11111111,
            type_=0x2222, subtype=0x3333,
            length=0x00000004,
        )

        output = record.serialize()

        eq_(buf, output)


class TestMrtlibOspf2MrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.Ospf2MrtRecord.
    """

    @mock.patch('ryu.lib.packet.ospf.ospf.parser')
    def test_parse(self, mock_ospf_parser):
        remote_ip = '10.0.0.1'
        local_ip = '10.0.0.2'
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0b\x00\x00'  # type=TYPE_OSPFv2(11), subtype
            b'\x00\x00\x00\x0c'  # length=len(remote_ip + local_ip + body)
            + addrconv.ipv4.text_to_bin(remote_ip)  # remote_ip
            + addrconv.ipv4.text_to_bin(local_ip)   # local_ip
            + body               # ospf_message
        )

        mock_ospf_message = mock.MagicMock(spec=ospf.OSPFMessage)
        mock_ospf_parser.return_value = (mock_ospf_message, None, '')

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(mrtlib.MrtRecord.TYPE_OSPFv2, record.type)
        eq_(0x0000, record.subtype)
        eq_(0x0000000c, record.length)
        eq_(remote_ip, record.message.remote_ip)
        eq_(local_ip, record.message.local_ip)
        eq_(mock_ospf_message, record.message.ospf_message)
        eq_(b'', rest)

    def test_serialize(self):
        remote_ip = '10.0.0.1'
        local_ip = '10.0.0.2'
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0b\x00\x00'  # type=TYPE_OSPFv2(11), subtype
            b'\x00\x00\x00\x0c'  # length=len(remote_ip + local_ip + body)
            + addrconv.ipv4.text_to_bin(remote_ip)  # remote_ip
            + addrconv.ipv4.text_to_bin(local_ip)   # local_ip
            + body               # ospf_message
        )

        mock_ospf_message = mock.MagicMock(spec=ospf.OSPFMessage)
        mock_ospf_message.serialize.return_value = body

        message = mrtlib.Ospf2MrtMessage(
            remote_ip=remote_ip,
            local_ip=local_ip,
            ospf_message=mock_ospf_message,
        )
        record = mrtlib.Ospf2MrtRecord(
            message=message,
            timestamp=0x11111111,
            # type_=None,
            # subtype=None,
            # length=None,
        )

        output = record.serialize()

        eq_(buf, output)


class TestMrtlibTableDumpMrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.TableDumpMrtRecord.
    """

    @mock.patch('ryu.lib.packet.bgp._PathAttribute.parser')
    def test_parse_afi_ipv4(self, mock_bgp_attr_parser):
        prefix = '10.0.0.0'
        peer_ip = '172.16.0.1'
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0c\x00\x01'  # type=TYPE_TABLE_DUMP(12),
                                 # subtype=SUBTYPE_AFI_IPv4(1)
            b'\x00\x00\x00\x1a'  # length=26
            b'\x22\x22\x33\x33'  # view_num, seq_num
            + addrconv.ipv4.text_to_bin(prefix) +  # prefix
            b'\x18\x01'          # prefix_len=24, status=1
            b'\x44\x44\x44\x44'  # originated_time
            + addrconv.ipv4.text_to_bin(peer_ip) +  # peer_ip
            b'\xfd\xe8\x00\x04'  # peer_as=65000, attr_len=len(body)
            + body               # bgp_attributes
        )

        mock_bgp_attr = mock.MagicMock(spec=bgp._PathAttribute)
        mock_bgp_attr_parser.return_value = (mock_bgp_attr, b'')

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(mrtlib.MrtRecord.TYPE_TABLE_DUMP, record.type)
        eq_(mrtlib.TableDumpMrtRecord.SUBTYPE_AFI_IPv4, record.subtype)
        eq_(0x0000001a, record.length)
        eq_(0x2222, record.message.view_num)
        eq_(0x3333, record.message.seq_num)
        eq_(prefix, record.message.prefix)
        eq_(24, record.message.prefix_len)
        eq_(1, record.message.status)
        eq_(0x44444444, record.message.originated_time)
        eq_(peer_ip, record.message.peer_ip)
        eq_(65000, record.message.peer_as)
        eq_(0x0004, record.message.attr_len)
        eq_([mock_bgp_attr], record.message.bgp_attributes)
        eq_(b'', rest)

    def test_serialize_afi_ipv4(self):
        prefix = '10.0.0.0'
        peer_ip = '172.16.0.1'
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0c\x00\x01'  # type=TYPE_TABLE_DUMP(12),
                                 # subtype=SUBTYPE_AFI_IPv4(1)
            b'\x00\x00\x00\x1a'  # length=26
            b'\x22\x22\x33\x33'  # view_num, seq_num
            + addrconv.ipv4.text_to_bin(prefix) +  # prefix
            b'\x18\x01'          # prefix_len=24, status=1
            b'\x44\x44\x44\x44'  # originated_time
            + addrconv.ipv4.text_to_bin(peer_ip) +  # peer_ip
            b'\xfd\xe8\x00\x04'  # peer_as=65000, attr_len=len(body)
            + body               # bgp_attributes
        )

        mock_bgp_attr = mock.MagicMock(spec=bgp._PathAttribute)
        mock_bgp_attr.serialize.return_value = body

        message = mrtlib.TableDumpAfiIPv4MrtMessage(
            view_num=0x2222,
            seq_num=0x3333,
            prefix=prefix,
            prefix_len=24,
            status=1,
            originated_time=0x44444444,
            peer_ip=peer_ip,
            peer_as=65000,
            bgp_attributes=[mock_bgp_attr],
            # attr_len=4
        )
        record = mrtlib.TableDumpMrtRecord(
            message=message,
            timestamp=0x11111111,
            # type_=None,
            # subtype=None,
            # length=None,
        )

        output = record.serialize()

        eq_(buf, output)

    @mock.patch('ryu.lib.packet.bgp._PathAttribute.parser')
    def test_parse_afi_ipv6(self, mock_bgp_attr_parser):
        prefix = '2001:db8::1'
        peer_ip = 'fe80::1'
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0c\x00\x02'  # type=TYPE_TABLE_DUMP(12),
                                 # subtype=SUBTYPE_AFI_IPv6(2)
            b'\x00\x00\x00\x32'  # length=50
            b'\x22\x22\x33\x33'  # view_num, seq_num
            + addrconv.ipv6.text_to_bin(prefix) +  # prefix
            b'\x40\x01'          # prefix_len=64, status=1
            b'\x44\x44\x44\x44'  # originated_time
            + addrconv.ipv6.text_to_bin(peer_ip) +  # peer_ip
            b'\xfd\xe8\x00\x04'  # peer_as=65000, attr_len=len(body)
            + body               # bgp_attributes
        )

        mock_bgp_attr = mock.MagicMock(spec=bgp._PathAttribute)
        mock_bgp_attr_parser.return_value = (mock_bgp_attr, b'')

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(mrtlib.MrtRecord.TYPE_TABLE_DUMP, record.type)
        eq_(mrtlib.TableDumpMrtRecord.SUBTYPE_AFI_IPv6, record.subtype)
        eq_(0x00000032, record.length)
        eq_(0x2222, record.message.view_num)
        eq_(0x3333, record.message.seq_num)
        eq_(prefix, record.message.prefix)
        eq_(64, record.message.prefix_len)
        eq_(1, record.message.status)
        eq_(0x44444444, record.message.originated_time)
        eq_(peer_ip, record.message.peer_ip)
        eq_(65000, record.message.peer_as)
        eq_(0x0004, record.message.attr_len)
        eq_([mock_bgp_attr], record.message.bgp_attributes)
        eq_(b'', rest)

    def test_serialize_afi_ipv6(self):
        prefix = '2001:db8::1'
        peer_ip = 'fe80::1'
        body = b'test'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0c\x00\x02'  # type=TYPE_TABLE_DUMP(12),
                                 # subtype=SUBTYPE_AFI_IPv6(2)
            b'\x00\x00\x00\x32'  # length=50
            b'\x22\x22\x33\x33'  # view_num, seq_num
            + addrconv.ipv6.text_to_bin(prefix) +  # prefix
            b'\x40\x01'          # prefix_len=64, status=1
            b'\x44\x44\x44\x44'  # originated_time
            + addrconv.ipv6.text_to_bin(peer_ip) +  # peer_ip
            b'\xfd\xe8\x00\x04'  # peer_as=65000, attr_len=len(body)
            + body               # bgp_attributes
        )

        mock_bgp_attr = mock.MagicMock(spec=bgp._PathAttribute)
        mock_bgp_attr.serialize.return_value = body

        message = mrtlib.TableDumpAfiIPv6MrtMessage(
            view_num=0x2222,
            seq_num=0x3333,
            prefix=prefix,
            prefix_len=64,
            status=1,
            originated_time=0x44444444,
            peer_ip=peer_ip,
            peer_as=65000,
            bgp_attributes=[mock_bgp_attr],
            # attr_len=4
        )
        record = mrtlib.TableDumpMrtRecord(
            message=message,
            timestamp=0x11111111,
            # type_=None,
            # subtype=None,
            # length=None,
        )

        output = record.serialize()

        eq_(buf, output)


class TestMrtlibTableDump2MrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.TableDump2MrtRecord.
    """

    # Note: The classes corresponding to the following subtypes are
    # tested in TestMrtlibMrtRecord.
    # - SUBTYPE_PEER_INDEX_TABLE = 1
    # - SUBTYPE_RIB_IPV4_UNICAST = 2
    # - SUBTYPE_RIB_IPV4_MULTICAST = 3
    # - SUBTYPE_RIB_IPV6_UNICAST = 4
    # - SUBTYPE_RIB_IPV6_MULTICAST = 5

    @mock.patch('ryu.lib.mrtlib.MrtRibEntry.parse')
    @mock.patch('ryu.lib.packet.bgp.BGPNLRI.parser')
    def test_parse_rib_generic(self, mock_nlri_parser, mock_rib_entry_parser):
        nlri_bin = b'nlri'  # 4 bytes
        rib_entries_bin = b'ribs'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0d\x00\x06'  # type=TYPE_TABLE_DUMP_V2(13),
                                 # subtype=SUBTYPE_RIB_GENERIC(6)
            b'\x00\x00\x00\x11'  # length=17
            b'\x22\x22\x22\x22'  # seq_num
            b'\x33\x33\x44'      # afi, safi
            + nlri_bin +         # nlri
            b'\x00\x01'          # entry_count
            + rib_entries_bin    # rib_entries
        )
        buf_entries = (
            b'\x00\x01'          # entry_count
            + rib_entries_bin    # rib_entries
        )

        mock_bgp_nlri = mock.MagicMock(spec=bgp._AddrPrefix)
        mock_nlri_parser.return_value = (mock_bgp_nlri, buf_entries)

        mock_rib_entry = mock.MagicMock(spec=mrtlib.MrtRibEntry)
        mock_rib_entry_parser.return_value = (mock_rib_entry, b'')

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(mrtlib.MrtRecord.TYPE_TABLE_DUMP_V2, record.type)
        eq_(mrtlib.TableDump2MrtRecord.SUBTYPE_RIB_GENERIC, record.subtype)
        eq_(0x00000011, record.length)
        eq_(0x22222222, record.message.seq_num)
        eq_(0x3333, record.message.afi)
        eq_(0x44, record.message.safi)
        eq_(mock_bgp_nlri, record.message.nlri)
        eq_(0x0001, record.message.entry_count)
        eq_([mock_rib_entry], record.message.rib_entries)
        eq_(b'', rest)

    def test_serialize_rib_generic(self):
        nlri_bin = b'nlri'  # 4 bytes
        rib_entries_bin = b'ribs'  # 4 bytes
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x0d\x00\x06'  # type=TYPE_TABLE_DUMP_V2(13),
                                 # subtype=SUBTYPE_RIB_GENERIC(6)
            b'\x00\x00\x00\x11'  # length=17
            b'\x22\x22\x22\x22'  # seq_num
            b'\x33\x33\x44'      # afi, safi
            + nlri_bin +         # nlri
            b'\x00\x01'          # entry_count
            + rib_entries_bin    # rib_entries
        )

        mock_bgp_nlri = mock.MagicMock(spec=bgp._AddrPrefix)
        mock_bgp_nlri.serialize.return_value = nlri_bin

        mock_rib_entry = mock.MagicMock(spec=mrtlib.MrtRibEntry)
        mock_rib_entry.serialize.return_value = rib_entries_bin

        message = mrtlib.TableDump2RibGenericMrtMessage(
            seq_num=0x22222222,
            afi=0x3333,
            safi=0x44,
            nlri=mock_bgp_nlri,
            rib_entries=[mock_rib_entry],
            # entry_count=1,
        )
        record = mrtlib.TableDump2MrtRecord(
            message=message,
            timestamp=0x11111111,
            # type_=None,
            # subtype=None,
            # length=None,
        )

        output = record.serialize()

        eq_(buf, output)


class TestMrtlibMrtPeer(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.MrtPeer.
    """

    def test_parse_two_octet_as(self):
        bgp_id = '1.1.1.1'
        ip_addr = '10.0.0.1'
        buf = (
            b'\x00'      # type
            + addrconv.ipv4.text_to_bin(bgp_id)     # bgp_id
            + addrconv.ipv4.text_to_bin(ip_addr) +  # ip_addr
            b'\xfd\xe8'  # as_num
        )

        peer, rest = mrtlib.MrtPeer.parse(buf)

        eq_(0, peer.type)
        eq_(bgp_id, peer.bgp_id)
        eq_(ip_addr, peer.ip_addr)
        eq_(65000, peer.as_num)
        eq_(b'', rest)

    def test_serialize_two_octet_as(self):
        bgp_id = '1.1.1.1'
        ip_addr = '10.0.0.1'
        buf = (
            b'\x00'      # type
            + addrconv.ipv4.text_to_bin(bgp_id)     # bgp_id
            + addrconv.ipv4.text_to_bin(ip_addr) +  # ip_addr
            b'\xfd\xe8'  # as_num
        )

        peer = mrtlib.MrtPeer(
            bgp_id=bgp_id,
            ip_addr=ip_addr,
            as_num=65000,
            # type_=0,
        )

        output = peer.serialize()

        eq_(buf, output)


class TestMrtlibBgp4MpMrtRecord(unittest.TestCase):
    """
    Test case for ryu.lib.mrtlib.Bgp4MpMrtRecord.
    """

    # Note: The classes corresponding to the following subtypes are
    # tested in TestMrtlibMrtRecord.
    # - SUBTYPE_BGP4MP_MESSAGE = 1
    # - SUBTYPE_BGP4MP_MESSAGE_AS4 = 4
    # - SUBTYPE_BGP4MP_STATE_CHANGE_AS4 = 5
    # - SUBTYPE_BGP4MP_MESSAGE_LOCAL = 6
    # - SUBTYPE_BGP4MP_MESSAGE_AS4_LOCAL = 7

    def test_parse_state_change_afi_ipv4(self):
        peer_ip = '10.0.0.1'
        local_ip = '10.0.0.2'
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x10\x00\x00'  # type=TYPE_BGP4MP(16),
                                 # subtype=SUBTYPE_BGP4MP_STATE_CHANGE(0)
            b'\x00\x00\x00\x14'  # length=20
            b'\xfd\xe9\xfd\xea'  # peer_as=65001, local_as=65002
            b'\x22\x22\x00\x01'  # if_index, addr_family=AFI_IPv4(1)
            + addrconv.ipv4.text_to_bin(peer_ip)     # peer_ip
            + addrconv.ipv4.text_to_bin(local_ip) +  # local_ip
            b'\x00\x01\x00\x02'  # old_state=STATE_IDLE(1),
                                 # new_state=STATE_CONNECT(2)
        )

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(mrtlib.MrtRecord.TYPE_BGP4MP, record.type)
        eq_(mrtlib.Bgp4MpMrtRecord.SUBTYPE_BGP4MP_STATE_CHANGE, record.subtype)
        eq_(0x00000014, record.length)
        eq_(65001, record.message.peer_as)
        eq_(65002, record.message.local_as)
        eq_(0x2222, record.message.if_index)
        eq_(mrtlib.Bgp4MpStateChangeMrtMessage.AFI_IPv4,
            record.message.afi)
        eq_(mrtlib.Bgp4MpStateChangeMrtMessage.STATE_IDLE,
            record.message.old_state)
        eq_(mrtlib.Bgp4MpStateChangeMrtMessage.STATE_CONNECT,
            record.message.new_state)
        eq_(b'', rest)

    def test_serialize_state_change_afi_ipv4(self):
        peer_ip = '10.0.0.1'
        local_ip = '10.0.0.2'
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x10\x00\x00'  # type=TYPE_BGP4MP(16),
                                 # subtype=SUBTYPE_BGP4MP_STATE_CHANGE(0)
            b'\x00\x00\x00\x14'  # length=20
            b'\xfd\xe9\xfd\xea'  # peer_as=65001, local_as=65002
            b'\x22\x22\x00\x01'  # if_index, addr_family=AFI_IPv4(1)
            + addrconv.ipv4.text_to_bin(peer_ip)     # peer_ip
            + addrconv.ipv4.text_to_bin(local_ip) +  # local_ip
            b'\x00\x01\x00\x02'  # old_state=STATE_IDLE(1),
                                 # new_state=STATE_CONNECT(2)
        )

        message = mrtlib.Bgp4MpStateChangeMrtMessage(
            peer_as=65001,
            local_as=65002,
            if_index=0x2222,
            peer_ip=peer_ip,
            local_ip=local_ip,
            old_state=mrtlib.Bgp4MpStateChangeMrtMessage.STATE_IDLE,
            new_state=mrtlib.Bgp4MpStateChangeMrtMessage.STATE_CONNECT,
            # afi=mrtlib.Bgp4MpStateChangeMrtMessage.AFI_IPv4,
        )
        record = mrtlib.Bgp4MpMrtRecord(
            message=message,
            timestamp=0x11111111,
            # type_=None,
            # subtype=None,
            # length=None,
        )

        output = record.serialize()

        eq_(buf, output)

    def test_parse_state_change_afi_ipv6(self):
        peer_ip = 'fe80::1'
        local_ip = 'fe80::2'
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x10\x00\x00'  # type=TYPE_BGP4MP(16),
                                 # subtype=SUBTYPE_BGP4MP_STATE_CHANGE(0)
            b'\x00\x00\x00\x2c'  # length=44
            b'\xfd\xe9\xfd\xea'  # peer_as=65001, local_as=65002
            b'\x22\x22\x00\x02'  # if_index, addr_family=AFI_IPv6(2)
            + addrconv.ipv6.text_to_bin(peer_ip)     # peer_ip
            + addrconv.ipv6.text_to_bin(local_ip) +  # local_ip
            b'\x00\x01\x00\x02'  # old_state=STATE_IDLE(1),
                                 # new_state=STATE_CONNECT(2)
        )

        (record, rest) = mrtlib.MrtRecord.parse(buf)

        eq_(0x11111111, record.timestamp)
        eq_(mrtlib.MrtRecord.TYPE_BGP4MP, record.type)
        eq_(mrtlib.Bgp4MpMrtRecord.SUBTYPE_BGP4MP_STATE_CHANGE, record.subtype)
        eq_(0x0000002c, record.length)
        eq_(65001, record.message.peer_as)
        eq_(65002, record.message.local_as)
        eq_(0x2222, record.message.if_index)
        eq_(mrtlib.Bgp4MpStateChangeMrtMessage.AFI_IPv6,
            record.message.afi)
        eq_(mrtlib.Bgp4MpStateChangeMrtMessage.STATE_IDLE,
            record.message.old_state)
        eq_(mrtlib.Bgp4MpStateChangeMrtMessage.STATE_CONNECT,
            record.message.new_state)
        eq_(b'', rest)

    def test_serialize_state_change_afi_ipv6(self):
        peer_ip = 'fe80::1'
        local_ip = 'fe80::2'
        buf = (
            b'\x11\x11\x11\x11'  # timestamp
            b'\x00\x10\x00\x00'  # type=TYPE_BGP4MP(16),
                                 # subtype=SUBTYPE_BGP4MP_STATE_CHANGE(0)
            b'\x00\x00\x00\x2c'  # length=44
            b'\xfd\xe9\xfd\xea'  # peer_as=65001, local_as=65002
            b'\x22\x22\x00\x02'  # if_index, addr_family=AFI_IPv6(2)
            + addrconv.ipv6.text_to_bin(peer_ip)     # peer_ip
            + addrconv.ipv6.text_to_bin(local_ip) +  # local_ip
            b'\x00\x01\x00\x02'  # old_state=STATE_IDLE(1),
                                 # new_state=STATE_CONNECT(2)
        )

        message = mrtlib.Bgp4MpStateChangeMrtMessage(
            peer_as=65001,
            local_as=65002,
            if_index=0x2222,
            peer_ip=peer_ip,
            local_ip=local_ip,
            old_state=mrtlib.Bgp4MpStateChangeMrtMessage.STATE_IDLE,
            new_state=mrtlib.Bgp4MpStateChangeMrtMessage.STATE_CONNECT,
            # afi=mrtlib.Bgp4MpStateChangeMrtMessage.AFI_IPv4,
        )
        record = mrtlib.Bgp4MpMrtRecord(
            message=message,
            timestamp=0x11111111,
            # type_=None,
            # subtype=None,
            # length=None,
        )

        output = record.serialize()

        eq_(buf, output)
